/*
 *  Main routine for the miniloader.
 *
 *  david.rusling@reo.mts.dec.com
 */
#include <linux/kernel.h>
#include <linux/config.h>
#include <linux/string.h>
#include <linux/version.h>
#include <linux/mm.h>
#include <linux/kernel_stat.h>
#include <linux/pci.h>
#include <linux/bios32.h>
#include <linux/version.h>
#include <linux/compile.h>

#include <asm/system.h>
#include <asm/io.h>
#include <asm/console.h>
#include <asm/hwrpb.h>
#include <asm/page.h>
#include <asm/pgtable.h>

#include <stdarg.h>
#include <filehdr.h>
#include <aouthdr.h>
#include <scnhdr.h>

#include "miniboot.h"
#include "impure.h"
#include "uart.h"

#ifndef NULL
#define NULL ((void *)0)
#endif

/*****************************************************************************
 *  Build Controlling Definitions.                                           *
 *****************************************************************************/
/*
 *  These are not in config.in as they'd clutter things up too much.  So, they
 *  are here.
 */
/* DEBUG_MINIBOOT: this turns on much more printk() statements.  Defining this
 * *may* help you find a problem */
#undef DEBUG_MINIBOOT 

/* DEBUG_USE_DBM: this allows you to use the Evaluation Board minidebugger to
 * step through the code. */
#undef DEBUG_USE_DBM  

/* FLOPPY_HACK: this means that code will explicitly go and set the floppy to a 
   known state before anything else happens */
#ifndef CONFIG_ALPHA_NONAME
#define FLOPPY_HACK 1
#endif

/* BOOT_VERBOSE: this means that you will get lots of messages as the system
   boots.  Handy for debugging, but not for anything else. */
#define BOOT_VERBOSE 

/* DEFAULT_MEMORY_SIZE: define this and we don't go look in the PALcode impure
   area to figure out the memory size */
#undef DEFAULT_MEMORY_SIZE 0x2000000

/*****************************************************************************
 *  Global variables.                                                        *
 *****************************************************************************/
/*
 *  For device drivers and so on.
 */
#ifndef MINI_LINUX_IN_MEMORY
int ramdisk_size;
#endif
/*
 *  memory allocation
 */
static char *memory_map = 0;
volatile unsigned int firstpfn = 0;
volatile unsigned int lastpfn = 0;

#ifndef MINI_LINUX_IN_MEMORY
static char *temp_memory_ptr = 0;
#endif

/*
 *  HWRPB
 */
struct hwrpb_struct *hwrpb;

/*
 *  memory mapping
 */
unsigned int ptbr = 0;
U64 vptbr = 0;
struct pcb_struct boot_pcb;

/*
 *  Someplace to put the arguments to the kernel.
 */
U64 *pkernel_args;
U64 kernel_at;
U64 kernel_entry;
U64 kernel_pages;

/*
 *  Counting stuff.
 */
unsigned int num_l2_pages = 0;
unsigned int num_l3_pages = 0;

struct hae hae = {
  0, 
  (unsigned long *) HAE_ADDRESS
  };


struct kernel_stat kstat;

/*
 *  What type of alpha system is this *realy*
 */
int alpha_sys_type;

/*****************************************************************************
 *  Macros.                                                                  *
 *****************************************************************************/
#ifndef TRUE 
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif

#define PASS1_LCA_TYPE	       0x100000004ULL
					/* Pass 1 LCA chip detect */

/*
 *  External stuff.
 */
extern int vga_initialised;
extern void __start_again(void);

/*
 *  Forward declarations.
 */
#ifndef MINI_LINUX_IN_MEMORY
char *malloc(int size);
#endif
void mark_many_pfn(int pfn, int count, int type);
void check_kernel_header(char *header);
void boot_main(void);
void boot_main_cont(void);
void build_HWRPB_mem_map(void);


#ifdef CONFIG_BLK_DEV_FD

#ifdef FLOPPY_HACK

/* fda */
#define FDADCR	0x03F2
#define FDAMSR	0x03F4
#define FDADR	0x03F5
#define FDADRR	0x03F7
#define FDADCB	0x03F7

/* dcr */
#define DCRSELA	0x00
#define DCRSELB	0x01
#define DCRSELC	0x02
#define DCRSELD	0x03
#define DCRNRES	0x04
#define DCRENAB	0x08
#define DCRMTRA	0x10
#define DCRMTRB	0x20
#define DCRMTRC	0x40
#define DCRMTRD	0x80

/* msr */
#define MSRDIO	0x40
#define MSRRQM	0x80

/* drr */
#define DRR500	0x00
#define DRR250	0x02

/* dcb */
#define DCBNCHG	0x80

/* st0 */
#define ST0IC	0xC0
#define ST0NT	0x00
#define ST0NR	0x08

/* st1 */
#define ST1NW	0x02

/* cmd */
#define NREAD	0x66
#define NWRITE	0x45
#define NRECAL	0x07
#define NSEEK	0x0F
#define NSINTS	0x08
#define NSPEC	0x03

#define FDTOMEM	0
#define MEMTOFD	1

#define NCMD	9
#define NSTS	7

#define UNITNUM	0
#define UNITSEL	(DCRSELA)
#define UNITMTR	(DCRMTRA)

#define lSRT	0xE0			/* 4 mS                          */
#define lHUT	0x08			/* 256 mS                        */

#define hSRT	0xC0			/* 4 mS                          */
#define hHUT	0x0F			/* 256 mS                        */

#define HLT	0x02			/* 4 mS                          */
#define ND	0x00			/* Use DMA                       */

#define SRT	0xE0			/* 4 mS                          */
#define HUT	0x08			/* 256 mS                        */
#define HLT	0x02			/* 4 mS                          */
#define ND	0x00			/* Use DMA                       */

void
  fdasleep(int nmsec)
{
    int i;

    while (nmsec--) {
	i = 50000;
	do {
	} while (--i);
    }
}

int
  fdasts(void) 
{
    int nsts;
    int byte;
    char sts[NSTS];

#ifdef DEBUG_MINIBOOT
    int i;

#endif

    nsts = 0;
    for (;;) {
	while (((byte = inb(FDAMSR)) & MSRRQM) == 0)
	    ;
	if ((byte & MSRDIO) == 0)
	    break;
	byte = inb(FDADR);
	if (nsts < NSTS)
	    sts[nsts++] = byte;
    }

#ifdef DEBUG_MINIBOOT
    printk("sts:");
    for (i = 0; i < nsts; ++i)
	printk(" %02x", sts[i] & 0xFF);
    printk("\n");
#endif

    if ((sts[0] & ST0IC) != ST0NT) {
	return (0);
    }
    return (1);
}

void fdacmd(char cmd[], int ncmd)
{
    int i;

#ifdef DEBUG_MINIBOOT
    printk("cmd:");
    for (i = 0; i < ncmd; ++i)
	printk(" %02x", cmd[i] & 0xFF);
    printk("\n");
#endif

    for (i = 0; i < ncmd; ++i) {
	while ((inb(FDAMSR) & (MSRRQM | MSRDIO)) != MSRRQM)
	    ;
	outb(cmd[i], FDADR);
    }
}

void floppy_init_hack(void) 
{
    char cmd[1];

    printk("hack alert! Initialising the floppy\n");

    outb(DCRENAB | DCRNRES, FDADCR);
    fdasleep(50);
    outb(DRR250, FDADRR);
    cmd[0] = NSINTS;
    fdacmd(cmd, 1);
    (void) fdasts();
    cmd[0] = NSPEC;
    cmd[1] = SRT | HUT;
    cmd[2] = HLT | ND;
    fdacmd(cmd, 3);
}
#endif

#endif

/*****************************************************************************
 *  Swap to a new PALcode                                                    *
 *****************************************************************************/
void swap_to_palcode(void *palbase, void *pc, void *ksp, int new_ptbr, 
  void *new_vptbr)
{
    extern void swppal(void *, void *, struct pcb_struct *, void *);
/*
 *  Set up the initial pcb.
 */

    boot_pcb.ksp = (U64) ksp;
    boot_pcb.usp = 0;
    boot_pcb.ptbr = new_ptbr;
    boot_pcb.pcc = 0;
    boot_pcb.asn = 0;
    boot_pcb.unique = 0;
    boot_pcb.flags = 1;

/*
 *  Call swppal.  The arguments are:
 *
 *  a0:      PAL address
 *  a1:      PC
 *  a2:      pcb
 *  a3:      vptbr
 */
#if 0
    halt();
#else
    swppal(palbase, pc, &boot_pcb, new_vptbr);
#endif
}

/*****************************************************************************
 *  Kernel Loading code                                                      *
 *****************************************************************************/
/*
 *  checks that the kernel header passed is valid.  Note that this code assumes
 *  that all sections of the header are in memory and contiguous with each other.
 */
void check_kernel_header(char *header)
{
    struct filehdr *fh;
    struct aouthdr *ah;
    struct scnhdr *sh;

    fh = (struct filehdr *) header;
    ah = (struct aouthdr *) (header + sizeof(struct filehdr));
    sh = (struct scnhdr *) (header
			    + sizeof(struct filehdr) + sizeof(struct aouthdr));

#ifdef BOOT_VERBOSE
    printk("...filehdr @0x%p, fh->f_opthdr = 0x%x, aouthdr @0x%p\n", 
      (void *) fh, fh->f_opthdr, (void *) ah);
    printk("   f_magic: 0x%x, f_nscns: 0x%x, f_timdat: 0x%x\n", 
      fh->f_magic, fh->f_nscns, fh->f_timdat);
    printk("   f_symptr: 0x%lX, f_nsyms: %d\n", fh->f_symptr, fh->f_nsyms);
    printk("   f_opthdr: 0x%x, f_flags: 0x%x\n", fh->f_opthdr, fh->f_flags);

    printk("...aouthdr fields:\n");
    printk("   magic: 0%o, vstamp: 0x%x, bldrev: 0x%x, padcell: 0x%x\n", 
      ah->magic, ah->vstamp, ah->bldrev, ah->padcell);
    printk("   text_start: %lx, ", ah->text_start);
    printk("   tsize: %lx\n", ah->tsize);
    printk("   data_start: %lx, ", ah->data_start);
    printk("   dsize: %lx\n", ah->dsize);
    printk("   bss_start: %lx, ", ah->bss_start);
    printk("   bsize: %lx\n", ah->bsize);
    printk("   entry: %lx\n", ah->entry);
    printk("   gprmask: 0x%x, fprmask: 0x%x, gp_value: 0x%lx\n", 
      ah->gprmask, ah->fprmask, 
      ah->gp_value);
#endif

    /* Do some sanity checks... */
    if (fh->f_magic != ALPHAMAGIC) {
	printk("Bad filehdr magic number 0x%x should be 0x%x\n", 
	  fh->f_magic, ALPHAMAGIC);
	halt();
    }
    if ((ah->magic != OMAGIC) && (ah->magic != ZMAGIC)) {
	printk("Bad aouthdr magic number 0x%x should be 0x%x\n", 
	  ah->magic, OMAGIC);
	halt();
    }
    if (ah->text_start != START_ADDR) {
	printk("Kernel must start at virtual address 0x%lX, not 0x%lX\n", 
	       START_ADDR, ah->text_start);
	halt();
    }
}
/*
 *  Returns the entry point of the kernel, or zero if kernel fails to load.
 */
U64 load_image_from_memory(char *address, unsigned int pages)
{
    struct filehdr *fh;
    struct aouthdr *ah;
    struct scnhdr *sh;

    int offset;

    int i;

    unsigned char *ptr;

    printk("Loading the Linux kernel from memory (at 0x%p)\n", address);
/*
 *  Get pointers to the sections - all in the first buffer read.
 */
    fh = (struct filehdr *) address;
    ah = (struct aouthdr *) (address + sizeof(struct filehdr));
    sh = (struct scnhdr *) (address
			    + sizeof(struct filehdr) + sizeof(struct aouthdr));

    /* check the kernel header */
    check_kernel_header((char *) address);

    /* save the entry point */
    kernel_at = ah->text_start;
    kernel_at |= IDENT_ADDR;
    kernel_entry = ah->entry;
    kernel_entry |= IDENT_ADDR;

    /* How many pages does the kernel take up?  Don't forget to include the
     * pages between KERNEL_START and START_ADDR */
    kernel_pages = 
      (ah->tsize + ah->dsize + ah->bsize + PAGE_SIZE - 1) >> PAGE_SHIFT;
    kernel_pages += 
      ((kernel_entry - KERNEL_START + PAGE_SIZE - 1) >> PAGE_SHIFT);

#ifdef BOOT_VERBOSE
    printk("Kernel uses %lld page(s)\n", kernel_pages);
    printk("   Text: 0x%lx - 0x%lx\n", ah->text_start, 
      ah->text_start + ah->tsize - 1);
    printk("   Data: 0x%lx - 0x%lx\n", ah->data_start, 
      ah->data_start + ah->dsize - 1);
    printk("   BSS: 0x%lx - 0x%lx\n", ah->bss_start, 
      ah->bss_start + ah->bsize - 1);
#endif

/*
 *  Copy the kernel.
 */
#ifdef DEBUG_MINIBOOT
    printk("...copying the Kernel into physical memory @ 0x%p\n", kernel_at);
#endif

    offset = sh->s_scnptr;
    memcpy((char *) kernel_at, address + offset, ah->tsize + ah->dsize);
/*
 *  Zero the bss part of the image.
 */
#ifdef DEBUG_MINIBOOT
    printk("...clearing bss memory between 0x%p and 0x%p\n", ah->bss_start, 
      (char *) ah->bss_start + ah->bsize);
#endif

    /* for some reason, memset does not seem to work in this case */
#if 0
    memset(ah->bss_start, 0, (char *) ah->bss_start + ah->bsize);
#else
    ptr = (unsigned char *) ah->bss_start;
    for (i = 0; i < ah->bsize; i++)
	*ptr++ = 0;
#endif

    return kernel_at;
}
#ifndef MINI_LINUX_IN_MEMORY
#define IOBUF_SIZE 1024
/*
 *  Returns the entry point of the kernel, or zero if kernel fails to load.
 */
U64 load_image_from_device(char *device, char *file)
{
    struct filehdr *fh;
    struct aouthdr *ah;
    struct scnhdr *sh;

    int offset;
    int count;

    int channel;
    int fd;
    int blkno;

    char *iobuf;
    char *to;

    unsigned int text_pages;

    unsigned int *ptr;
    char *cptr;

    int zipped;

/*
 * We must read the linux kernel off of the disk.
 */

#ifdef BOOT_VERBOSE
    printk("Searching for %s on device %s\n", file, device);
#endif

    channel = device_mount(device);
    if (channel)
	return 0;

/*
 *  Look for the file
 */
    fd = ext2_open(file);
#ifdef BOOT_VERBOSE
    printk("...%s\n", fd >= 0 ? "found it" : "failed to find it");
#endif

    if (fd < 0)
	return 0;

/*
 *  Is it a zipped image (.gz suffix)?
 */
    zipped = FALSE;
    cptr = file + strlen(file) - 3;
    if (strcmp((const char *) cptr, (const char *) ".gz") == 0)
	zipped = TRUE;
    else {
	if (strcmp((const char *) cptr, (const char *) ".GZ") == 0)
	    zipped = TRUE;
    }
    if (zipped) {
	char *where;
/*
 *  Uncompress it into the end of memory.
 */

#ifdef BOOT_VERBOSE
	printk("Marking Linux in memory as occupied (TEMP): ");
#endif
	mark_many_pfn(lastpfn - _PHYSICAL_TO_PFN(START_SIZE),
	  _PHYSICAL_TO_PFN(START_SIZE), 
	  TEMPORARY_PAGE);
	where =
	  (char *) _PFN_TO_PHYSICAL(lastpfn - _PHYSICAL_TO_PFN(START_SIZE));
	uncompress_kernel(fd, where);
	kernel_entry = load_image_from_memory((char *) where,
	  _PHYSICAL_TO_PFN(START_SIZE));
#ifdef BOOT_VERBOSE
	printk("...freeing up memory occupied by the kernel: ");
#endif
	mark_many_pfn(_PHYSICAL_TO_PFN(where), _PHYSICAL_TO_PFN(START_SIZE),
	  FREE_PAGE);
    } else {
/*
 *  Allocate an iobuf for the transfers and read the first block of the
 *  file into it (the section headers).  Make it 512 bytes aligned.
 */
	iobuf = malloc(IOBUF_SIZE + 512) + 512;
	iobuf = (char *) ((unsigned long) iobuf & (~(512 - 1)));
#ifdef DEBUG_MINIBOOT
	printk("...allocated iobuf at 0x%p\n", iobuf);
#endif
	blkno = 0;			/* logical block number */
	ext2_bread(fd, blkno++, iobuf);

/*
 *  Get pointers to the sections - all in the first buffer read.
 */
	fh = (struct filehdr *) iobuf;
	ah = (struct aouthdr *) (iobuf + sizeof(struct filehdr));
	sh = (struct scnhdr *) (iobuf + sizeof(struct filehdr)
				+ sizeof(struct aouthdr));

/*
 *  Check out the header
 */
	check_kernel_header((char *) iobuf);

	text_pages = (ah->tsize + PAGE_SIZE - 1) >> PAGE_SHIFT;

	/* save the entry point */
	kernel_entry = ah->entry;
	kernel_entry |= IDENT_ADDR;
	kernel_at = ah->text_start;
	kernel_at |= IDENT_ADDR;

	/* How many pages does the kernel take up?  Don't forget to include the
	 * pages between KERNEL_START and START_ADDR */
	kernel_pages = 
	  (ah->tsize + ah->dsize + ah->bsize + PAGE_SIZE - 1) >> PAGE_SHIFT;
	kernel_pages += ((kernel_entry - KERNEL_START) >> PAGE_SHIFT);


#ifdef BOOT_VERBOSE
	printk("First block loaded; kernel uses %lld page(s)\n", kernel_pages);
	printk("   Text: 0x%lx - 0x%lx\n", ah->text_start, 
	  ah->text_start + ah->tsize - 1);
	printk("   Data: 0x%lx - 0x%lx\n", ah->data_start, 
	  ah->data_start + ah->dsize - 1);
	printk("   BSS: 0x%lx - 0x%lx\n", ah->bss_start, 
	  ah->bss_start + ah->bsize - 1);
#endif

/*
 *  Copy the kernel bit by bit from the file into memory.
 */
	printk("...copying %s into memory (block size = 0x%X)", 
	  file, IOBUF_SIZE);
	printk(" at physical address 0x%lX\n", ah->text_start);
	offset = sh->s_scnptr;
#ifdef BOOT_VERBOSE
	printk("...Text segment offset @0x%x (%d)\n", offset, offset);
#endif
/*
 *  We may have the first part of the image in this block.  Check before
 *  we embark on a reading frenzy.
 */
	count = 0;
	to = (char *) ah->text_start;

	/* read the first part of the image from the iobuf */
	if (offset < IOBUF_SIZE) {
	    int size;

	    size = IOBUF_SIZE - offset;
	    memcpy(kernel_at, iobuf + offset, size);
	    printk("#");

	    count += size;
	    to += size;
	    offset = IOBUF_SIZE;
	}

	/* read the rest of the blocks directly where we want them */
	while (count < (ah->tsize + ah->dsize)) {
	    ext2_bread(fd, blkno++, to);
#if 1
	    printk("#");
#else
	    printk("%08X\r", to);
#endif
	    to += IOBUF_SIZE;
	    count += IOBUF_SIZE;
	}
	printk("\n");
/*
 *  Zero the bss part of the image.
 */
#ifdef DEBUG_MINIBOOT
	printk("...clearing bss memory between 0x%p and 0x%p\n", 
	  ah->bss_start, 
	  (char *) ah->bss_start + ah->bsize);
#endif

	/* for some reason, memset does not seem to work in this case */
#if 0
	memset(ah->bss_start, 0, (char *) ah->bss_start + ah->bsize);
#endif
	ptr = (unsigned int *) ah->bss_start;
	for (count = 0; count < ah->bsize / sizeof(unsigned int); count++)
	    *ptr++ = 0;

	ext2_close(channel);

    }
#ifdef BOOT_VERBOSE
    printk("...Linux kernel loaded into memory, entry @ 0x%llx\n",
	   kernel_entry);
#endif
    return kernel_entry;
}

#endif

/*****************************************************************************
 *  Load Linux (from wherever)                                               *
 *****************************************************************************/
/*
 *  The linux kernel is either on one of the known devices or it is linked
 *  into this image at a known location.
 *  
 *  If MINI_LINUX_IN_MEMORY is defined and the user boots from device
 * "embedded:", then we assume an embedded linux image.
 */

void boot_cmd(char *devname, char *filename, char *boot_string)
{
    U64 entry;
    U64 *kernel_args;
    unsigned int pfn;
    char *page;


#ifdef MINI_LINUX_IN_MEMORY
    if (kernel_at) {
	entry = load_image_from_memory((char *) kernel_at, kernel_pages);
    }
    else {
	printk("MILO: No embedded kernel to boot\n");
	entry = 0;
    }
#else
    entry = load_image_from_device(devname, filename);
#endif

    if (entry == 0) {
	printk("MILO: Failed to load the kernel\n");
	return;
    }

    /* now we've loaded the kernel, mark the memory that it occupies as in use
     */
#ifdef BOOT_VERBOSE
    printk("...freeing original marked kernel pages: ");
#endif
    mark_many_pfn(_PHYSICAL_TO_PFN(KERNEL_START), 
      _PHYSICAL_TO_PFN(START_SIZE), 
      FREE_PAGE);
#ifdef BOOT_VERBOSE
    printk("...marking real kernel pages as occupied: ");
#endif
    mark_many_pfn(_PHYSICAL_TO_PFN(KERNEL_START), kernel_pages, 
      ALLOCATED_PAGE);


/*****************************************************************************
 *  Build the final memory map                                               *
 *****************************************************************************/

    /* mark as free any pages currently marked as TEMPORARY_PAGE */
#ifdef BOOT_VERBOSE
    printk("...freeing temporary pages\n");
#endif
    page = memory_map;
    for (pfn = 0; pfn < lastpfn + 1; pfn++) {
	if (*page == TEMPORARY_PAGE) *page = FREE_PAGE;
	page++;
    }

    /* This must be done when all required memory has been allocated */
    build_HWRPB_mem_map();

/*****************************************************************************
 *  Set up the area where we pass any boot arguments.                        *
 *****************************************************************************/

    /* for now, just make sure that it is zero */
    memset((void *) ZERO_PGE, 0, PAGE_SIZE);

    if (strlen(boot_string) == 0) {
	if (getenv("BOOT_STRING")) {
	    strcat((char *) ZERO_PGE, getenv("BOOT_STRING"));
	} else {
	    strcat((char *) ZERO_PGE, "root=/dev/sda2");
	}
    } else
	strcat((char *) ZERO_PGE, boot_string);

    printk("Boot line is %s\n", (char *) ZERO_PGE);

/*****************************************************************************
 *  Set up the Kernel calling information block                              *
 *****************************************************************************/
/*
 *  This is located below this image in memory.  It is 7*8 bytes long and
 *  contains the following:
 * 
 *  Offset      Contents
 *    0*8       Linux kernel entry point
 */
    pkernel_args = kernel_args = (U64*) ((U64) LOADER_AT + 8);
    kernel_args[0] = (U64)entry;

    printk("Bootstrap complete, kernel @ 0x%lx, sp @ 0x%lx\n", entry, 
	   INIT_STACK + PAGE_SIZE);

/*****************************************************************************
 *  Swap to the new page tables and jump to the Linux Kernel                 *
 *****************************************************************************/

    /* change context */
    printk("...turning on virtual addressing and jumping to the Linux Kernel\n");

    /* build the pcb */
    boot_pcb.ksp = INIT_STACK + PAGE_SIZE;
    boot_pcb.usp = 0;
    boot_pcb.ptbr = ptbr;
    boot_pcb.pcc = 0;
    boot_pcb.asn = 0;
    boot_pcb.unique = 0;
    boot_pcb.flags = 1;

    /* bump up the ipl */
    cli();

    /* now switch to the new scheme of things. */
    printk("\n------------------------------------------------------------\n\n");
    swap_to_palcode((void *) PALCODE_AT, 
      (void *) entry, (void *) INIT_STACK + PAGE_SIZE, 
      ptbr, (void *) vptbr);
/*
 *  If we ever get to this point, something has gone seriously wrong!!!
 */
    printk("If you can read this, something went VERY wrong!\n");
#ifdef NEVER
    halt();
#endif

}

#ifndef MINI_LINUX_IN_MEMORY
/* Load a standalone executable file from disk and run it.  This is 
 * different from boot in that we don't have to set up the system to run
 * Linux.  At the moment, updateflash is the only such executable.
 */
int run_cmd(char *devname, char *filename)
{
    U64 entry;

    entry = load_image_from_device(devname, filename);
    if (entry) {
	void (*routine_address)(void);

	/* call the image that we have just loaded. */
	routine_address = (void (*) (void))entry;
	(routine_address)();

	/* If we get here, then I guess we can return to the command loop...
	 * although there's no guarantees as to what the program may have done 
	 * to our state... */
	return (0);
    }
    return (-1);
}
#endif

/*****************************************************************************
 *  Memory allocation code                                                   *
 *****************************************************************************/
#ifndef MINI_LINUX_IN_MEMORY
#undef  GUARD_BANDS 
/*
 *  Allocate some space that we can free up later.
 */
char *malloc(int size)
{
    char *result;
    int top, bottom, pages, i, room, pfn;
    char *mm;

    size += 8 - (size & 0x7);		/* align to an 8 byte boundary */
#ifdef GUARD_BANDS
    size += (4 * 8);
#endif
    pages = (size + PAGE_SIZE) / PAGE_SIZE;

    bottom = _PHYSICAL_TO_PFN(temp_memory_ptr);
    top = _PHYSICAL_TO_PFN(temp_memory_ptr + size);

    /* is there already some memory that we might be able to use? */
    if (temp_memory_ptr != NULL) {
	room = TRUE;
	mm = (char *)memory_map + bottom + 1; 
	for (i = bottom; i < top; i++) {
	    if (*mm++ != FREE_PAGE) {
		room = FALSE;
		break;
	    }
	}
	if (!room) {
	    /* no room, allocate more memory */
	    pfn = allocate_many_pfn(pages, TEMPORARY_PAGE);
	    temp_memory_ptr = _PFN_TO_PHYSICAL(pfn);
	}
    } else {
	/* nope, allocate some */
	pfn = allocate_many_pfn(pages, TEMPORARY_PAGE);
	temp_memory_ptr = (char *) _PFN_TO_PHYSICAL(pfn);
    }

    /* ok, did we get or have enough memory? */
    if (temp_memory_ptr != NULL) {
	result = temp_memory_ptr;
	temp_memory_ptr += size;
    } else {
	result = NULL;
    }

#ifdef DEBUG_MINIBOOT
    printk("malloc(): returning 0x%p\n", result);
#endif
    memset(result, 0, size);
#ifdef GUARD_BANDS
    {
    U64 *guard = (U64 *)result;

    *guard++ = 0xAAAABBBBCCCCDDDDUL;
    *guard =   0xAAAABBBBCCCCDDDDUL;

    guard = result + size - 16;
    *guard++ = 0x1111222233334444UL;
    *guard =   0x1111222233334444UL;

    result = result + 16;

    }
#endif
    return result;
}

void free(void *where)
{
/*
 * for now, do nothing
 */
}

#endif
/*
 *  Allocate many free pages.
 */
unsigned int allocate_many_pfn(unsigned int count, int type)
{
    unsigned int pfn, i;
    char *first = memory_map;
    char *last;

#ifdef DEBUG_MINIBOOT
    printk("allocate_many_pfn: called, count = %d\n", count);
#endif

    for (pfn = 0; pfn < lastpfn + 1; pfn++) {
	if (*first == FREE_PAGE) {

	    /* we've found the first page, are there enough free ones together?
	     */
	    if (pfn + count > lastpfn + 1) {
		printk("...Fatal error: not enough memory\n");
		halt();
	    } else {
		int found = 1;

		last = first;
		for (i = 0; i < count; i++) {
		    if (*last != FREE_PAGE) {
			found = 0;
			break;
		    }
		    last++;
		}
		if (found) {
		    for (i = 0; i < count; i++)
			*first++ = type;
#ifdef DEBUG_MINIBOOT
		    printk("allocate_many_pfn: returning %d\n", pfn);
#endif
		    return pfn;
		}
	    }
	}
	first++;
    }
    printk("...Fatal error: not enough memory\n");
    halt();
    return 0;
}
void mark_many_pfn(int pfn, int count, int type)
{
    char *page;
    int i;

#ifdef BOOT_VERBOSE
    printk("(pfn %04d to %04d)\n", pfn, pfn + count - 1);
#endif
    page = memory_map + pfn;
    for (i = 0; i < count; i++)
	*page++ = type;
}

/*
 *  Allocate a free page.
 */
unsigned int allocate_pfn(int type)
{
    unsigned int pfn;
    char *page;

    page = memory_map;
    for (pfn = 0; pfn < lastpfn + 1; pfn++) {
#if 0
#ifdef DEBUG_MINIBOOT
	printk("pfn = 0x%x, usage = %d\n", pfn, *page);
#endif
#endif
	if (*page == FREE_PAGE) {
	    *page = type;
	    return pfn;
	}
	page++;
    }
    printk("...Fatal error: not enough memory\n");
    halt();
    return 0;
}

void zeropage_phys(int pfn)
{
#if 1
    unsigned long *where;
    int size, i;

    where = (unsigned long *) _PFN_TO_PHYSICAL(pfn);
    size = PAGE_SIZE / sizeof(unsigned long);
    for (i = 0; i < size; i++)
	*where++ = 0;
#endif

}
void free_pfn(int pfn)
{
    char *page;

    printk("(pfn %04d)\n", pfn);
    page = memory_map + pfn;
    *page = FREE_PAGE;
}
/*
 *  Initialise the memory allocation scheme.  This is called early on.  It assumes
 *  that the miniloader is loaded at PALCODE_AT (PALcode first, followed by an image) 
 *  and that the stack is below the image in memory.
 *  
 *  Memory is marked as:
 *
 *          0 = free
 *          1 = free, but nonvolatile
 *          2 = reserved to console
 */
void init_mem(U64 Top)
{

    printk("Initialising the memory\n");
/*
 *  Take the top page for the memory map.
 */
    memory_map = (char *) (PAGE_OFFSET | (Top - PAGE_SIZE));
    lastpfn = _PHYSICAL_TO_PFN(memory_map);
#ifdef BOOT_VERBOSE
    printk("...Last pfn is %04d (used to hold memory map)\n", lastpfn);
#endif
    firstpfn = 0;
#ifdef BOOT_VERBOSE
    printk("...memory map @ 0x%x\n", memory_map);
#endif
/*
 *  Initialise the memory map.
 */

    /* mark everything as free (initially) */
#ifdef BOOT_VERBOSE
    printk("...mark all pages as free: ");
#endif
    mark_many_pfn(0, lastpfn + 1, FREE_PAGE);

    /* mark the memory map as not free */
#ifdef BOOT_VERBOSE
    printk("...marking page containing memory map as occupied: ");
#endif
    mark_many_pfn(lastpfn, 1, TEMPORARY_PAGE);

    /* mark the ZERO_PGE as not free */
#ifdef BOOT_VERBOSE
    printk("...marking ZERO_PGE as occupied: ");
#endif
    mark_many_pfn(_PHYSICAL_TO_PFN(ZERO_PGE), 1, ALLOCATED_PAGE);

    /* mark the INIT_STACK as not free */
#ifdef BOOT_VERBOSE
    printk("...marking INIT_STACK as occupied: ");
#endif
    mark_many_pfn(_PHYSICAL_TO_PFN(INIT_STACK), 1, ALLOCATED_PAGE);

    /* REMOVE THIS IF YOU WANT MORE MEMORY, LEAVE IT IF YOU WANT TO RUN THE
     * MINIDEBUGGER */
#ifdef DEBUG_USE_DBM

    /* for now mark the area occupied by the Debug Monitor palcode as not free
     */
#ifdef BOOT_VERBOSE
    printk("...(TEMP) marking Debug Monitor PALcode as occupied: ");
#endif
    mark_many_pfn(0, _PHYSICAL_TO_PFN(PALCODE_SIZE), ALLOCATED_PAGE);

    /* for now, mark the area occupied by the debug monitor as not free */
#ifdef BOOT_VERBOSE
    printk("...(TEMP) marking Debug Monitor as occupied: ");
#endif
    mark_many_pfn(_PHYSICAL_TO_PFN(PALCODE_SIZE), _PHYSICAL_TO_PFN(0x100000), 
      ALLOCATED_PAGE);

#endif

    /* for now, mark the area occupied by the miniloader as not free */
#ifdef BOOT_VERBOSE
    printk("...(TEMP) marking Miniloader as occupied: ");
#endif
    mark_many_pfn(_PHYSICAL_TO_PFN(LOADER_AT), _PHYSICAL_TO_PFN(LOADER_SIZE), 
      TEMPORARY_PAGE);

    /* mark the memory that will be occupied by the kernel as reserved.  The
     * number of pages is a guess; put it right when we read in the image */
#ifdef BOOT_VERBOSE
    printk("...marking Linux kernel as occupied (TEMP): ");
#endif
    mark_many_pfn(_PHYSICAL_TO_PFN(KERNEL_START), 
      _PHYSICAL_TO_PFN(START_SIZE), 
      TEMPORARY_PAGE);

    /* mark the area occupied by the palcode as not free */
#ifdef BOOT_VERBOSE
    printk("...marking PALcode as occupied: ");
#endif
    mark_many_pfn(_PHYSICAL_TO_PFN(PALCODE_AT), 
      _PHYSICAL_TO_PFN(PALCODE_SIZE), 
      ALLOCATED_PAGE);

#ifdef MINI_LINUX_IN_MEMORY

    /* if the kernel is in memory, then we move it to the top of memory
     * (because it may overlap with where it needs to really go in memory) */
    do {
	int size;
	AOUTHDR *ah;
	SCNHDR *sh;

	/* Work out just how big the image is (in pages). */
	ah = (AOUTHDR *) (LINUX_AT + sizeof(struct filehdr));
	sh = (SCNHDR *) (LINUX_AT + sizeof(struct filehdr)
			 + sizeof(struct aouthdr));
	size = ah->tsize + ah->dsize + sh->s_scnptr;
	kernel_pages = (size + PAGE_SIZE - 1) >> PAGE_SHIFT;

	kernel_at = (U64) _PFN_TO_PHYSICAL(lastpfn - kernel_pages);
#ifdef BOOT_VERBOSE
	printk("...moving the in-memory Linux kernel to 0x%p\n", kernel_at);
#endif
	memcpy(kernel_at, (char *) LINUX_AT, size);
#ifdef BOOT_VERBOSE
	printk("...marking Linux in memory as occupied: ");
#endif
	mark_many_pfn(_PHYSICAL_TO_PFN(kernel_at), kernel_pages, 
	  TEMPORARY_PAGE);
    } while (0);
#endif

}
/*****************************************************************************
 *  Virtual address mapping (page table) code.                               *
 *****************************************************************************/
/*
 *  Given all of the information, build a valid PTE.
 */
pte_t build_pte(unsigned long pfn, pgprot_t protection)
{
    pte_t pte;

    pte_val(pte) = pte_val(mk_pte(_PFN_TO_PHYSICAL(pfn), protection));
    pte_val(pte) = pte_val(pte) | _PAGE_VALID;

#ifdef DEBUG_MINIBOOT
    printk("build_pte: pte = 0x%X,0x%X\n", pte_val(pte) >> 32, pte_val(pte));
#endif
    return pte;
}

/*
 *  Given a VA, make a set of page table entries for it (if they
 *  don't already exist).
 *
 *  Args:
 *
 *  va            virtual address to add page table entries for.
 *  vpfn          the PFN that the virtual address is to be mapped to.
 *  protection    the protection to set the L3 page table entry to (all L1 and
 *                L2 pages are marked as Kernel read/write).
 */
void add(U64 va, unsigned int vpfn, unsigned int protection)
{
    U64 l1, l2, l3;
    unsigned int pfn, newpfn;
    U64 pa;
    pte_t pte;

/*
 *  Dismantle the VA.
 */

    l1 = 0;
    l2 = (va & 0xff800000) >> 20;
    l3 = (va & 0x007fe000) >> 10;

#ifdef DEBUG_MINIBOOT
    printk("...Adding PTEs for virtual address 0x%X, pfn = %X\n", va, vpfn);
    printk("...L1 offset = 0x%x, L2 offset = 0x%x, L3 offset = 0x%x\n", 
      l1, l2, l3);
#endif
/*
 *  Now build a set of entries for it (if they do not already
 *  exist).  In the next blocks of code, pfn is always the pfn of
 *  the current entry and newpfn is the pfn that is referenced from
 *  it.
 */
/*
 *  L1.  The L1 PT PFN is held in the global integer ptbr. 
 */
    pa = _PFN_TO_PHYSICAL(ptbr) + l1;
    if (ReadQ(pa) == 0) {
	newpfn = _ALLOCATE_PFN();	/* for L2 */
	num_l2_pages++;
#ifdef DEBUG_MINIBOOT
	printk("\tAllocated L2 page table at PFN 0x%X (physical = 0x%X)\n", 
	  newpfn, _PFN_TO_PHYSICAL(newpfn));
#endif
	zeropage_phys(newpfn);
	pte = build_pte(newpfn, __pgprot(_PAGE_KWE | _PAGE_KRE));
	WriteQ(pa, pte_val(pte));
    } else
	pte_val(pte) = ReadQ(pa);
#ifdef DEBUG_MINIBOOT
    printk("L1 PTE at PFN(0x%x) + 0x%04X = 0x%X\n", 
      ptbr, l1, ReadQ(pa));
#endif
    pfn = _PHYSICAL_TO_PFN(pte_page(pte));

/*
 *  L2, pfn inherited from L1 entry (which may have been allocated
 *  above).
 */
    pa = _PFN_TO_PHYSICAL(pfn) + l2;
    if (ReadQ(pa) == 0) {
	num_l3_pages++;
	newpfn = _ALLOCATE_PFN();
#ifdef DEBUG_MINIBOOT
	printk("\tAllocated L3 page table at PFN 0x%08X (physical = 0x%X)\n", 
	  newpfn, _PFN_TO_PHYSICAL(newpfn));
#endif
	zeropage_phys(newpfn);
	pte = build_pte(newpfn, __pgprot(_PAGE_KWE | _PAGE_KRE));
	WriteQ(pa, pte_val(pte));
    } else
	pte_val(pte) = ReadQ(pa);
#ifdef DEBUG_MINIBOOT
    printk("L2 PTE at PFN(0x%x) + 0x%04X = 0x%08X\n", 
      pfn, l2, ReadQ(pa));
#endif
    pfn = _PHYSICAL_TO_PFN(pte_page((pte_t) pte));

/*
 *  L3, pfn inherited from L2 entry above.  vpfn contains the page frame
 *  number of the real thing whose PTEs we have just invented above.
 */
    pa = _PFN_TO_PHYSICAL(pfn) + l3;
    if (ReadQ(pa) == 0) {
	pte = build_pte(vpfn, __pgprot(protection));
	WriteQ(pa, pte_val(pte));
    } else
	pte_val(pte) = ReadQ(pa);
#ifdef DEBUG_MINIBOOT
    printk("L3 PTE at PFN(0x%x) + 0x%04X = 0x%08X\n", 
      pfn, l3, ReadQ(pa));
#endif
}

/*
 *  System specific routine that returns the top of memory (ie the
 *  amount of physical memory in the system).
 */
#ifdef DEFAULT_MEMORY_SIZE
#define memory_size() (u64)(DEFAULT_MEMORY_SIZE)
#else
#define CSERVE_K_RD_IMPURE     0x0B

U64 memory_size(void) 
{
    U64 PalImpureBase;
    U64 size;

#ifdef DEBUG_MINIBOOT
    printk("memory_size(), called\n");
#endif
#if 0
/*
 *  HACK ALERT! EB66 mem size is at 368 and the EB64+ is at 350
 */
#ifdef MINI_EB64
    size = 0x2000000;
#else
    PalImpureBase = (U64)(CNS_Q_BASE + cServe(0, 0, CSERVE_K_RD_IMPURE));
    size = (unsigned int) ReadQ(PalImpureBase + (U64) CNS_Q_MEM_SIZE);
    if (size == 0)
	size = 0x2000000;		/* HACK */
#endif

#else
    size = 0x1000000;
#endif
#ifdef DEBUG_MINIBOOT
    printk("memory_size(), returning %d\n", size);
#endif
    return size;
}
#endif

/*****************************************************************************
 *  Build the Page Tables                                                    *
 *****************************************************************************/
/*
 *  We are running with mapping 1-to-1 physical.  Before we can turn on memory
 *  mapping we must set up the PTEs that we need.
 *
 *  We need memory allocation/mapping set up before we can do this.
 */
void build_mm(void) 
{
    pte_t pte;

    printk("Building the Page Tables\n");

    /* use the same pfn that Linux will use later (SWAPPER_PGD (0x300000)).
     * This is the first page in the kernel (KERNEL_START).  It saves us
     * allocating a page and then telling Linux that it isn't available. */

    ptbr = _PHYSICAL_TO_PFN(SWAPPER_PGD);
    zeropage_phys(ptbr);
#ifdef BOOT_VERBOSE
    printk("...L1 PT (ptbr) at PFN 0x%X, physical address 0x%llX\n",
	   ptbr, _PFN_TO_PHYSICAL(ptbr));
#endif
/*
 *  The virtual page table base pointer is used as a shortcut to get to the
 *  L3 entry for any given faulting address.  It is a special entry in the L1
 *  page table.  This part is important.   Linux 1.2 sets up the vptbr so that
 *  the *last* entry in the L1 PT contains the L1 pfn.  I honour that here.
 *  Change this at your peril.
 */
#ifdef BOOT_VERBOSE
    printk("...setting up the virtual page table base register\n");
#endif
    vptbr = 0xfffffffe00000000UL;
    pte = build_pte(ptbr, __pgprot(_PAGE_KWE | _PAGE_KRE));
    WriteQ(_PFN_TO_PHYSICAL(ptbr) + (sizeof(pte) * 0x3ff), pte_val(pte));

/*
 *  Just for now, add the debug monitor and this code in as virtual addresses
 *  matching their physical addresses.
 */
#if 0
    do {
	unsigned long va;
	int pfn;

	printk("...Temporarily mapping in debug monitor\n");
	va = 0;
	pfn = 0;
	while (va < ((LOADER_AT + LOADER_SIZE) & 0xffffffff)) {
	    add(va, pfn, (_PAGE_KWE | _PAGE_KRE));
	    pfn++;
	    va += PAGE_SIZE;
	}
    } while (0);
#endif
}
/*****************************************************************************
 *  HWRPB routines                                                           *
 *****************************************************************************/
void init_HWRPB(void) 
{
#ifdef CONFIG_ALPHA_LCA
    struct percpu_struct *percpu;

#endif
    unsigned int pfn;

    printk("Building a HWRPB (part I)\n");

    /* allocate it */
    pfn = _ALLOCATE_PFN();
    zeropage_phys(pfn);

    add((U64) INIT_HWRPB, pfn, (_PAGE_KWE | _PAGE_KRE));
#ifdef BOOT_VERBOSE
    printk("...Hwrpb at virtual address 0x%p (pfn = 0x%x)\n", INIT_HWRPB, pfn);
#endif

    /* fill it out the basic bits */

    hwrpb = (struct hwrpb_struct *) _PFN_TO_PHYSICAL(pfn);
    hwrpb->phys_addr = (U64) hwrpb & 0xffffffff;
    hwrpb->size = sizeof(struct hwrpb_struct);

/*
 *  Set up the system type.  This is based on configure-time #defines 
 */

    /* wierd i/o subsystems */
#ifdef CONFIG_ALPHA_EB64
    hwrpb->sys_type = 18;
#endif

    /* apecs/21064 i/o */
#ifdef CONFIG_ALPHA_APECS
#ifdef CONFIG_ALPHA_EB64P
    hwrpb->sys_type = ST_DEC_EB64P;
#endif
#ifdef CONFIG_ALPHA_CABRIOLET
    hwrpb->sys_type = ST_DEC_EBPC64;
#endif
#endif

    /* 21066/68 style i/o */
#ifdef CONFIG_ALPHA_LCA
#ifdef CONFIG_ALPHA_EB66
    hwrpb->sys_type = ST_DEC_EB66;
#endif
#ifdef CONFIG_ALPHA_EB66P
    hwrpb->sys_type = ST_DEC_EB66P;
#endif
#ifdef CONFIG_ALPHA_NONAME
    hwrpb->sys_type = ST_DEC_AXPPCI_33;
#endif
    hwrpb->processor_offset = 2 * sizeof(struct hwrpb_struct);
    percpu = 
      (struct percpu_struct *) ((char *) hwrpb + hwrpb->processor_offset);
    percpu->variation = ~PASS1_LCA_TYPE;
#endif

    alpha_sys_type = hwrpb->sys_type;
#ifdef BOOT_VERBOSE
    printk("...sys_type = %ld\n", hwrpb->sys_type);
#endif
    hwrpb->intr_freq = 977 * 4096;


}

void build_HWRPB_mem_map(void) 
{
    struct memclust_struct *cluster;
    struct memdesc_struct *memdesc;
    char *page;
    unsigned int pfn, i;

    printk("Building HWRPB (part II)\n");
    memdesc = 
      (struct memdesc_struct *) ((char *) hwrpb + sizeof(struct hwrpb_struct));

/*
 *  build clusters describing the clumps of memory that are available.
 *
 *  usage = 0 = free
 *          1 = free, but nonvolatile
 *          2 = reserved to console
 */
#ifdef DEBUG_MINIBOOT
    printk("...memdesc @ 0x%x\n", memdesc);
#endif
    hwrpb->mddt_offset = (char *) memdesc - (char *) hwrpb;
    memdesc->numclusters = 0;
    cluster = (struct memclust_struct *) (&memdesc->cluster[0]);

/*
 *  Build the memory clusters from the memory map.
 */
    page = memory_map;
    memdesc->numclusters = 1;
    cluster->start_pfn = 0;
    cluster->usage = *page++;
    cluster->numpages = 1;
    pfn = 1;
    while (pfn < lastpfn + 1) {

	/* find the next free cluster */
	if (*page == cluster->usage) {
	    cluster->numpages++;
	} else {

	    /* new cluster */
	    cluster = (struct memclust_struct *) 
	      (&memdesc->cluster[memdesc->numclusters++]);
	    cluster->usage = *page;
	    cluster->start_pfn = pfn;
	    cluster->numpages = 1;
	}
	page++;
	pfn++;
    }

/*
 *  Now dump out the clusters.
 */
    printk("...Memory Clusters\n");
    for (i = 0; i < memdesc->numclusters; i++) {
	cluster = (struct memclust_struct *) (&memdesc->cluster[i]);
	printk("     Cluster %d, start pfn = %04ld, size = %ld, %s\n", i, 
	  cluster->start_pfn, cluster->numpages, 
	  cluster->usage == FREE_PAGE ? "free" : "reserved");
    }

}

/*****************************************************************************
 *  Utility routines                                                         *
 *****************************************************************************/
#define NAMES 13
static char *devnames[NAMES] = {"hda", "hdb", "hdc", "hdd", "sda", "sdb", 
  "sdc", "sdd", "sde", "fd", "xda", "xdb", NULL};
static int devnums[NAMES] = {0x300, 0x340, 0x1600, 0x1640, 0x800, 0x810, 
  0x820, 0x830, 0x840, 0x200, 0xD00, 0xD40, 0};

void parse_device(unsigned int device, char *name)
{
    int i;
    int number;

#ifdef DEBUG_MINIBOOT
    printk("parse_device: device is 0x%04x\n", device);
#endif

    /* start of someplace */
    strcpy(name, "/dev/");

    /* drop of the trailing number */
    number = device & 0xf;
    device = device & 0xfff0;

    /* look for the device */
    for (i = 0; i < NAMES; i++) {
	if (device == devnums[i]) {

	    /* we've found it */
	    strcat(name, devnames[i]);

	    if (number != 0) {
		int len = strlen(name);

		name[len] = '0' + number;
		name[len + 1] = '\0';
	    }

#ifdef DEBUG_MINIBOOT
	    printk("...returning %s\n", name);
#endif
	    return;
	}
    }

    strcat(name, "sda0");
#ifdef DEBUG_MINIBOOT
    printk("...returning %s\n", name);
#endif
}

int parse_device_name(char *name)
{
    int i;
    int number, len, dev;

    /* get the true length of the name */
    len = strlen(name);
    number = name[len - 1] - '0';
    if ((number >= 0) && (number <= 9))
	len = len - 1;
    else
	number = -1;

    /* compare it to the list of devices */
    for (i = 0; i < NAMES; i++) {
	if (strncmp(devnames[i], name, len) == 0) {

	    /* we've found it */
	    dev = devnums[i];
	    if (number != -1)
		dev += number;
#ifdef DEBUG_MINIBOOT
	    printk("parse_device_name()...returning 0x%04x\n", dev);
#endif
	    return dev;
	}
    }
    return -1;
}

/*****************************************************************************
 *  MILO Command processing.                                                 *
 *****************************************************************************/
static void print_help(void) 
{
    printk("MILO command summary:\n\n");

#ifdef MINI_LINUX_IN_MEMORY
    printk("  boot             - Boot Linux from memory\n");
#else
    printk("  boot [dev:file] [boot string]\n");
    printk("                   - Boot Linux from the specified device and file\n");
    printk("                   - and pass the boot string supplied\n");
    printk("  run dev:file     - Run the standalone program in dev:file\n");
#endif
    printk("  setenv VAR VALUE - Set the variable VAR to the specified VALUE\n");
    printk("  unsetenv VAR     - Delete the specified environment variable\n");
    printk("  printenv         - Display the current environment variable settings\n");
    printk("  show             - Display all known devices\n");
    printk("  help             - Print this text\n\n");

    printk("Devices are specified as: fd0, hda, sda...\n");
    printk("Environment variables that MILO cares about:\n");
    printk("  MEMORY_SIZE      - System memory size in megabytes\n");
    printk("  AUTOBOOT         - If set, MILO attempts to boot on powerup\n");
    printk("                     and enters command loop only on failure\n");
    printk("  AUTOBOOT_TIMEOUT - Seconds to wait before auto-booting on powerup\n");
    printk("  BOOT_DEV         - Specifies the default boot device\n");
    printk("  BOOT_FILE        - Specifies the default boot file\n");
    printk("  BOOT_STRING      - Specifies the boot string to pass to the kernel\n");
}

void milo(void) 
{

    /* If AUTOBOOT is set and we have something to boot from, then do so here. 
     * If AUTOBOOT_TIMEOUT is set, give the user a chance to bail out first...
     */

    printk("Linux/Alpha Miniloader (MILO) %s\n", UTS_VERSION);
    printk("Waiting 10 seconds for VGA to come up...\n");

    kbd_getc_with_timeout(10);

    if (getenv("AUTOBOOT")) {
	char *default_bootdev;
	char *default_bootfile;
	int autoboot_timeout = 0;

	default_bootdev = getenv("BOOT_DEV");
	default_bootfile = getenv("BOOT_FILE");

	if (getenv("AUTOBOOT_TIMEOUT")) {
	    autoboot_timeout = atoi(getenv("AUTOBOOT_TIMEOUT"));
	}

	if (default_bootdev && default_bootfile) {

	    /* We've got something to try.  Do the timeout if necessary... */
	    printk("MILO: About to boot %s:%s.\n", default_bootdev, 
	      default_bootfile);
	    if (autoboot_timeout) {
		printk("Hit any key to stop autoboot and go to command mode\n");
		printk("Seconds remaining: ");
		while (autoboot_timeout--) {
		    printk("%4d    \b\b\b\b\b\b\b\b", autoboot_timeout);
		    if (kbd_getc_with_timeout(1) >= 0) {
			printk("\nEntering command mode.\n");
			goto command_loop;
		    }
		}
	    }

	    /* If boot_cmd returns, then the boot failed and we fall through to
	     * the command loop. */
	    boot_cmd(default_bootdev, default_bootfile, "");
	}
    }


command_loop: 

    /* Main command loop. */
    for (;;) {
	char command_string[256];
	char *cmd;

	printk("MILO> ");
	kbd_gets(command_string, 256);

	/* Grab the command word */
	cmd = strtok(command_string, " \t");

	/* We only support a few commands:
	 * boot [dev:file]         - Boot
	 * Linux      run dev:file            - Load and run a standalone
	 * executable.      setenv var [value]      - Set an environment
	 * variable.      unsetenv var            - Delete an environment
	 * variable.      printenv                - Print all currently-defined
	 *                                environment variables. */
	if (strcmp(cmd, "boot") == 0) {
	    char *bootdev;
	    char *bootfile;

	    /* Skip to the next token */
	    cmd = strtok(NULL, " \t");

	    /* If no device/file is specified and we don't have defaults set,
	     * inform the user. */
	    if (cmd == NULL) {
#ifndef MINI_LINUX_IN_MEMORY
		bootdev = getenv("BOOT_DEV");
		bootfile = getenv("BOOT_FILE");
		if ((bootdev == NULL) || (bootfile == NULL)) {
		    printk("MILO: 'boot' with no args requires that the BOOT_DEV and\n");
		    printk("       BOOT_FILE environment variables both be set.\n");
		    continue;
		}
#endif
		/* If we return from boot_cmd, it means we have to go back
		 * around the loop 'cause the boot failed! */
		boot_cmd(bootdev, bootfile, "");
		continue;
	    }
	      else {

#ifndef MINI_LINUX_IN_MEMORY
		  char *bootstring;

		/* The boot spec consists of a device spec and a file spec
		 * separated by a colon or whitespace. */
		bootstring = strtok(NULL, """");
		bootdev = strtok(cmd, ": \t");
		bootfile = strtok(NULL, " \t");
		if ((bootdev == NULL) || (bootfile == NULL)) {
		    printk("MILO: 'boot' command requires a device name and a file name\n");
		    printk("      separated by a colon or whitespace, e.g.\n");
		    printk("      MILO> boot fd0:vmlinux\n");
		    continue;
		}
#endif
		/* If we return from boot_cmd, it means we have to go back
		 * around the loop 'cause the boot failed! */
		if (strlen(bootstring) != 0) {
		    if (bootstring[0] == '"') bootstring[0] = ' ';
		    if (bootstring[strlen(bootstring) - 1] == '"') 
			bootstring[strlen(bootstring) - 1] = ' ';
		}
		boot_cmd(bootdev, bootfile, bootstring);
		continue;
	    }
	}
#ifndef MINI_LINUX_IN_MEMORY
	  else if (strcmp(cmd, "run") == 0) {
	    char *rundev;
	    char *runfile;

	    /* "run" is similar to "boot" except it requires the dev:file at
	     * all times (no defaults). */
	    cmd = strtok(NULL, " \t");

	    /* The run spec consists of a device spec and a file spec separated
	     * by a colon or whitespace. */
	    rundev = strtok(cmd, ": \t");
	    runfile = strtok(NULL, " \t");
	    if ((rundev == NULL) || (runfile == NULL)) {
		printk("MILO: 'run' command requires a device name and a file name\n");
		printk("      separated by a colon or whitespace, e.g.\n");
		printk("      MILO> run fd0:updateflash\n");
		continue;
	    }

	    /* If we return from run_cmd, it means we have to go back around
	     * the loop 'cause the load failed! */
	    run_cmd(rundev, runfile);
	    continue;
	}
#endif
	  else if (strcmp(cmd, "setenv") == 0) {
	    char envname[64];
	    char value[128];

	    /* Pluck off the environment variable name */
	    cmd = strtok(NULL, " \t");
	    if (cmd == NULL) {
		printk("MILO: setenv requires at least one argument\n");
		continue;
	    }

	    strcpy(envname, cmd);

	    /* Pluck off the value... */
	    value[0] = '\0';
	    cmd = strtok(NULL, " \t'\"");
	    if (cmd) {
		strcpy(value, cmd);
		while (cmd = strtok(NULL, " \t'\"")) {
		    strcat(value, " ");
		    strcat(value, cmd);
		}
	    }

	    /* do it! */
	    setenv(envname, value);
	}
	  else if (strcmp(cmd, "unsetenv") == 0) {
	    char envname[64];

	    /* Pluck off the environment variable name */
	    cmd = strtok(NULL, " \t");
	    if (cmd == NULL) {
		printk("MILO: unsetenv requires at least one argument\n");
		continue;
	    }
	    strcpy(envname, cmd);

	    cmd = strtok(NULL, " \t");
	    if (cmd) {
		printk("MILO: Too many arguments to unsetenv\n");
		continue;
	    }

	    unsetenv(envname);
	}
	  else if (strcmp(cmd, "printenv") == 0) {

	    /* See if there are any args... */
	    cmd = strtok(NULL, " \t");
	    if (cmd) {
		printk("MILO: No arguments allowed on printenv command\n");
		continue;
	    }

	    printenv();
	}
	  else if (strcmp(cmd, "help") == 0) {
	    print_help();
	}
	  else if (strcmp(cmd, "show") == 0) {
	    show_devices();
	}
	  else if (cmd) {
	    printk("MILO: Invalid command: %s\n", cmd);
	    print_help();
	    continue;
	}
    }

}
/*****************************************************************************
 *   Main entry point for system primary bootstrap code                      *
 *****************************************************************************/
/* 
 *
 * At this point we have been loaded into memory at LOADER_AT and we are running in Kernel
 * mode.  The Kernal stack base is at the top of memory (wherever that is).  Assume that it
 * takes up two 8K pages.  We are running 1-to-1 memory mapping (ie physical).
 */
void boot_main(void) 
{

/*****************************************************************************
 *  Initialize.                                                              *
 *****************************************************************************/
/*
 *  First initialize the Evaulation Board environment.
 */

    /* let's get explicit about some variables we care about. */
    vga_initialised = 0;
    num_l2_pages = num_l3_pages = 0;
    set_hae(hae.cache);

    /* we can only output to the serial port *after* we initialise it */
    uart_init();

#ifdef FLOPPY_HACK

    /* hack */
    floppy_init_hack();
#endif

    /* change the sp to be under the palcode minus a quad */
    wrsp(PALCODE_AT);
    wrfen();

#ifdef BOOT_VERBOSE
    printk("...memory size is 0x%X\n", memory_size());
#endif
/*
 *  Raise the IPL so that no interrupts can occur
 */
#ifdef BOOT_VERBOSE
    printk("...raising ipl from %d to 7\n", ipl(7));
#else
    ipl(7);
#endif

    /* initialise vga, after this vga messages will appear on the screen  */
#ifdef BOOT_VERBOSE
    printk("Initialising VGA\n");
#endif
#if 1 /* XXX davidm (can't invoke vga_init() before enabling VGA devices on PCI bus) */
    vga_init();
#endif

/*****************************************************************************
 *  Announce ourselves                                                       *
 *****************************************************************************/

#ifdef BOOT_VERBOSE
    printk("\nLinux/Alpha Evaluation Board bootstrap\n");
    printk("  version %s\n", UTS_VERSION);
    printk("  built on %s by %s\n", LINUX_COMPILE_HOST, LINUX_COMPILE_BY);
    printk("  compiler version %s\n", LINUX_COMPILER);
    printk("\n");
#endif

/*****************************************************************************
 *  Initialise the memory map                                                *
 *****************************************************************************/
/*
 *  Initialise the memory map.
 */
#ifdef DEBUG_MINIBOOT
    printk("Calling init_mem()\n");
#endif
    init_mem(memory_size());

/*****************************************************************************
 *  Building the page tables.                                                *
 *****************************************************************************/

    /* we cannot do this until we've set up the memory mapping and are able to 
     * allocate memory for page tables */
    build_mm();

/*****************************************************************************
 *  Build the basic HWRPB (needed by imported kernel code)                   *
 *****************************************************************************/

    /* we can't do this before the page tables are built */
    init_HWRPB();

/*****************************************************************************
 *  Swap to the new PALcode                                                  *
 *****************************************************************************/

    /* if we're trying to debug the code, then we cannot swap to a new PALcode.
     *   So, we just carry straight on.  We'll swap to the new PALcode when we 
     * start up Linux */

#ifdef DEBUG_USE_DBM
    boot_main_cont();
#else

    /* swap to the PALcode in this image, but stay in this image and make it
     * 1-to-1 virtual to physical address mapping */
    printk("Swapping to the new PALcode (1-to-1)\n");
    swap_to_palcode((void *) PALCODE_AT, (void *) __start_again, 
      (void *) (INIT_STACK + PAGE_SIZE), 
      (void *) 0, (void *) 1);
#endif
}

/*****************************************************************************
 *   Secondary entry point for system primary bootstrap code                 *
 *****************************************************************************/
/* 
 *
 *  At this point we have set up the memory map and swapped to the new PALcode.
 *  VGA is initialised and ipl is at 7.
 */
void boot_main_cont(void) 
{
/*****************************************************************************
 *  Do any CPU specific initialisation required                              *
 *****************************************************************************/

    /* we need the hwrpb set up before we can do this */
    cpu_init();

/*****************************************************************************
 *  Startup any device drivers required                                      *
 *****************************************************************************/

    /* we need the hwrpb set up before we can do this */
    device_init();

/*****************************************************************************
 *  Initialise the environment variable stuff                                *
 *****************************************************************************/
    env_init();

/*****************************************************************************
 *  Startup the keyboard so that we can get some input.                      *
 *****************************************************************************/
    milo();

/*
 *  If we ever get to this point, something has gone seriously wrong!!!
 */
    printk("If you can read this, something went VERY wrong!\n");
#ifdef NEVER
    halt();
#endif

}

void miniboot_panic(char *where, const char *fmt, ...)
{
    printk("panic called (@ 0x%p)\n", where - 4);
    halt();
}

/* We don't have this in our library, so I put it here... */
int atoi(char *s)
{
    int retval = 0;

    while (*s) {
	if ((*s < '0') || (*s > '9')) {
	    return (0);
	}
	retval *= 10;
	retval += (*s - '0');
	s++;
    }

    return (retval);
}

void doinitdata(void) 
{
#if 1
    extern unsigned long long _edata, _end;

    memset(&_edata, 0, (unsigned int) &_end - (unsigned int) &_edata);
#endif
}









