/*
 *  devices.c.   This part of the miniloader contains the ersatz (as in fake
 *  rollexs) Linux kernel.
 *
 *  david.rusling@reo.mts.dec.com.
 */
#include <linux/config.h>
#include <linux/fs.h>
#include <linux/timer.h>
#include <linux/signal.h>
#include <linux/errno.h>
#include <linux/pci.h>
#include <linux/tqueue.h>
#include <linux/sched.h>
#include <linux/locks.h>
#include <linux/major.h>
#include <linux/blkdev.h>
#include <linux/sched.h>
#include <linux/proc_fs.h>
#include <linux/major.h>
#include <linux/string.h>
#include <asm/hwrpb.h>
#include <asm/page.h>
#include <asm/unaligned.h>
#include "miniboot.h"
#include "uart.h"


#if 1
#define DEBUG_DEVICES 1
#endif

#define PCI_DEVICE_ID_SATURN            0x0484

#ifdef MINI_LINUX_IN_MEMORY
/*****************************************************************************
 * Initialise any devices                                                    *
 *****************************************************************************/
void device_init(void)
{ 

/*
 *  init
 */
    printk("Initializing timers\n");
    init_timers();

    printk("Initializing IRQs\n");
    init_IRQ();

    /* initialise the keyboard */
    printk("Initializing keyboard\n");
    kbd_init();

}

void show_devices(void)
{
    printk("Devices:\n    embedded\n");
}
#else LINUX_AT
/*
 *  Globals
 */
#ifdef CONFIG_BLK_DEV_FD
#define MAX_BUFFER_SECTORS 18
#endif

struct dummy {
    char dummy;
};

extern struct blk_dev_struct blk_dev[MAX_BLKDEV];

struct device_struct {
	const char * name;
	struct file_operations * fops;
};

struct device_struct blkdevs[MAX_BLKDEV] = {
        {NULL, NULL},
};


extern int *blk_size[];
extern int *blksize_size[];

int need_resched = 0;

struct wait_queue *wait_for_request;


dev_t ROOT_DEV = 0;
extern struct gendisk *gendisk_head;

unsigned long intr_count = 0;

#undef INIT_MM
#define INIT_MM { \
		0, \
		0, 0, 0, \
		0, 0, 0, 0, \
		0, 0, 0, 0, \
		0, \
/* ?_flt */	0, 0, 0, 0, \
		0, \
/* swap */	0, 0, 0, 0, \
		NULL, NULL }

mem_map_t * mem_map = NULL;

/*
 * for (at least) one IDE driver
 */
char dummy1;
struct drive_info_struct { char dummy[32]; } drive_info;
static struct sigaction init_sigaction[32] = { {0,}, };
static struct fs_struct init_fs = INIT_FS;
static struct files_struct init_files = INIT_FILES;
static struct mm_struct init_mm = INIT_MM;

/*
 *  INIT_TASK is used to set up the first task table, touch at
 * your own risk!. Base=0, limit=0x1fffff (=2MB)
 */
#define MINI_INIT_TASK \
/* state etc */	{ 0,15*HZ/100,15*HZ/100,0,0,0,0, \
/* debugregs */ { 0, },            \
/* exec domain */ NULL, \
/* binfmt */	NULL, \
/* schedlink */	&init_task,&init_task,&init_task,&init_task,\
/* signals */	init_sigaction, \
/* stack */	0,(unsigned long) NULL, \
/* ec,brk... */	0,0,0,0,0, \
/* pid etc.. */	0,0,0,0,0, \
/* suppl grps*/ {NOGROUP,}, \
/* proc links*/ &init_task,&init_task,NULL,NULL,NULL,NULL, \
/* uid etc */	0,0,0,0,0,0,0,0, \
/* timeout */	0,0,0,0,0,0,0, \
/* timer */	{ NULL, NULL, 0, 0, NULL }, \
/* utime */	0,0,0,0,0, \
/* rlimits */   { {LONG_MAX, LONG_MAX}, {LONG_MAX, LONG_MAX},  \
		  {LONG_MAX, LONG_MAX}, {LONG_MAX, LONG_MAX},  \
		  {       0, LONG_MAX}, {LONG_MAX, LONG_MAX}, \
		  {MAX_TASKS_PER_USER, MAX_TASKS_PER_USER}, {NR_OPEN, NR_OPEN}}, \
/* math */	0, \
/* comm */	"swapper", \
/* fs info */	0,NULL, \
/* ipc */	NULL, NULL, \
/* ldt */	NULL, \
/* tss */	INIT_TSS, \
/* fs */	&init_fs, \
/* files */	&init_files, \
/* mm */	&init_mm \
}

struct task_struct init_task = MINI_INIT_TASK;
struct task_struct *task[NR_TASKS];
struct task_struct *current = &init_task;

void *memory_map;

extern char *floppy_track_buffer;

#ifdef CONFIG_SCSI
void *dispatch_scsi_info_ptr = NULL;
struct proc_dir_entry scsi_dir[];
struct proc_dir_entry scsi_hba_dir[];
#endif
/*
 * Align VAL to ALIGN, which must be a power of two.
 */
#define ALIGN(val,align)	(((val) + ((align) - 1)) & ~((align) - 1))

/*
 *  This module implements the dummy device layer in the miniloader.
 *  It must be run after the PCI initialization code is run.
 */
extern char *malloc(unsigned int size);
extern void entUna(void);
extern void device_setup(void);
extern long blk_dev_init(long mem_start, long mem_end);

unsigned int current_device = 0;

/*****************************************************************************
 * Initialise any devices                                                    *
 *****************************************************************************/
void device_init(void)
{ 


    unsigned int i;
    unsigned char *start_mem, *end_mem;

/*
 *  init
 */
    for (i = 0; i < NR_TASKS; i++) task[i] = 0;
    task[0] = &init_task;
    current = task[0];

    printk("Initializing unaligned access traps\n");
    wrent(entUna, 4);

    printk("Initializing timers\n");
    init_timers();

    printk("Initializing IRQs\n");
    init_IRQ();

    printk("Initializing devices\n");

/*
 *  Get some temporary memory.
 */
    start_mem = malloc(8 * PAGE_SIZE);
    end_mem = start_mem + (8 * PAGE_SIZE);
#ifdef DEBUG_DEVICES
    printk("...start_mem @ 0x%p, end_mem @ 0x%p\n", start_mem, end_mem);
#endif

/*
 *  If we are setting up PCI, then do so.   
 */
#ifdef CONFIG_PCI

#ifdef CONFIG_ALPHA_LCA
#ifdef DEBUG_DEVICES
    printk("...LCA specific initialization\n");
#endif
    start_mem = (unsigned char *)lca_init((unsigned long)start_mem, (unsigned long)end_mem);
#endif
#ifdef CONFIG_ALPHA_APECS
#ifdef DEBUG_DEVICES
    printk("...APECS specific initialization\n");
#endif
    start_mem = (unsigned char *)apecs_init((unsigned long)start_mem, (unsigned long)end_mem);
#endif

    printk("Configuring PCI\n");
    start_mem = (unsigned char *)pci_init((unsigned long)start_mem, (unsigned long)end_mem);
#endif

#if 0 /* XXX davidm (can't invoke vga_init() before enabling VGA devices on PCI bus) */
    vga_init();
#endif
    for (i= 0; i < MAX_BLKDEV; i++) blkdevs[i].name = NULL;

/*
 *
 */
    current->pid = 0;
    current->blocked = 0;
    current->signal = 1;
/* 
 *  call the per device init routine.
 */

    printk("Configuring block devices\n");
    sti();
#ifdef PREDEFINED_ROOT_DEV
    ROOT_DEV = PREDEFINED_ROOT_DEV;
#else
    ROOT_DEV = 0x0800;		/* sda2 */
#endif
    start_mem = (unsigned char *)blk_dev_init((long) start_mem, (long) end_mem);
# ifdef DEBUG_DEVICES
    printk("after blk_dev_init: start_mem=%p\n", start_mem);
# endif

#ifdef CONFIG_SCSI
    printk("Configuring SCSI\n");
    sti();
    start_mem = scsi_dev_init(start_mem, end_mem);
# ifdef DEBUG_DEVICES
    printk("after scsi_dev_init: start_mem=%p\n", start_mem);
# endif
#endif

    if (start_mem >= end_mem) {
	panic("out of memory in device initialization");
    }

    device_setup();          /* in genhd.c */
/*
 *  Now tell the world what devices there are.
 */
    printk("Major devices are: ");
    for (i = 0; i < MAX_BLKDEV; i++) {
	if (blkdevs[i].name != NULL) 
	    printk("%s (%04X) ", blkdevs[i].name, i);
    }
    printk("\n");

    /* initialise the keyboard */
    printk("Initializing keyboard\n");
    kbd_init();


}

void show_devices(void)
{
    int i;

    printk("Devices:\n");
    for (i = 0; i < MAX_BLKDEV; i++) {
	if (blkdevs[i].name != NULL) 
	    printk("    %s (%04X) ", blkdevs[i].name, i << 8);
    }
    printk("\n");

}

/*****************************************************************************
 * Device routines                                                           *
 *****************************************************************************/

/*
 *  This called from boot main to mount the device so that it can get an
 *  image from it.
 */
static struct file file;
static struct inode inode;

int device_mount(char *device)
{
    int number, major, minor, result;

#ifdef DEBUG_DEVICES
    printk("device_mount(): called for %s\n", device);
#endif

    /* parse the device name into a device number, -1 means that we failed */
    number = parse_device_name(device);
    if (number == -1) return -1;

    /* carry on, major is the index intothe blkdevs[] structure */
    major = MAJOR(number);
    minor = MINOR(number);

    /* get the set of things that this device can do for us */
    file.f_op = blkdevs[major].fops;

/*
 *  Fill out the static structures that describe to the block device what we
 *  have opened.
 */
    file.f_pos = 0x0;     
    file.f_inode = &inode;
    file.f_mode = MS_RDONLY;
    inode.i_rdev = number;

/*
 *  If there's a device specific open routine, then go call it.
 *  The two device open routines that I've checked (floppy_open() and sd_open())
 *  both return 0 for success and -ERROR for failure.
 */
    if (file.f_op->open != NULL) {
	result = (file.f_op->open)(&inode, &file);
	if (result) {
	    printk("ERROR: failed to open device 0x%04X (result = %d)\n", device, result);
	    return -1;         /* error */
	}
    }

/*
 *  Now mount the filesystem.  Treat floppies differently from other harddisks.
 */
    if (major != FLOPPY_MAJOR) {
	struct gendisk *p;

	for (p = gendisk_head ; p ; p=p->next) {
	    if (major == p->major) {
		int partition_offset = 0;

		/* try and "mount" this partition */
		partition_offset = p->part[minor].start_sect;
		return ext2_mount(number); 
	    }
	}
	/* we goofed, return an error */
	printk("ERROR: trying to mount a non-existant device\n");
	return -1;
    } else 
	return ext2_mount(number);

    return 0;

}

/*
 *  Attempt to read an opened device.  fd is the fd returned by device_open()
 *  and is an index into the blkdevs[] array.
 */
int device_read(int fd, char *buffer, int size, int offset)
{
#ifdef DEBUG_DEVICES
    printk("device_read: called, %d bytes at offset 0x%X into buffer @ 0x%p\n", size, offset, buffer);
#endif

    if (file.f_op == NULL) {
	printk("...invalid device (f_op = NULL)\n");
	return 0;   /* invalid device */
    }

/*
 *  if there is one, call the read routine.
 */
    if (file.f_op->read != NULL) {
	int count;
#ifdef DEBUG_DEVICES
	int i;
#endif

	file.f_pos = offset;
	count = (file.f_op->read)(&inode, &file, buffer, size); 
	if (count < 0) {
	    printk("device_read: failed (%d)\n", count);
	    count = 0;
	} else {

#ifdef DEBUG_DEVICES		
	    printk("device_read: %d bytes read\n", count);
	    for (i=0; i<20; i++) printk("0x%02x ", buffer[i] & 0xFF);
	    printk("\n");
#endif
	}
	return count;
    } else {
	printk("...no read routine for this device\n");
	return 0;
    }

}

void device_close(int fd)
{
    printk("device_close: called, current_device = 0x%04X\n", current_device);
#ifdef DEBUG_DEVICES
    printk("device_close: called\n");
#endif
    inode.i_rdev = current_device;
    printk("device_close: called, release is @0x%p\n", file.f_op->release);
    if (file.f_op->release) (file.f_op->release)(&inode, &file);
/*  if (file.f_op->revalidate) (file.f_op->revalidate)(current_device); */
    current_device = -1;
}
/*****************************************************************************
 * Support routines                                                          *
 *****************************************************************************/

static inline void __sleep_on(struct wait_queue **p, int state)
{
    int *ptr = (int *)p;

#ifdef DEBUG_DEVICES
    printk("__sleep_on: called, queue = 0x%x, *p = %d\n", p, *p);
#endif

    *ptr = *ptr + 1;
/*
 *  Wait 100us.
 */
    while (*ptr > 0) udelay(10000);
}

void interruptible_sleep_on(struct wait_queue **p)
{
    unsigned long flags;

    save_flags(flags);
    sti();
#ifdef DEBUG_DEVICES		
    printk("interruptible_sleep_on: called, queue = 0x%x, ipl = %d\n", p, flags);
#endif
    __sleep_on(p,TASK_INTERRUPTIBLE);
    restore_flags(flags);
}

void sleep_on(struct wait_queue ** p)
{
#ifdef DEBUG_DEVICES		
    printk("sleep_on called for queue @ 0x%x\n", p);
#endif
}

void wake_up(struct wait_queue ** p)
{
    int *ptr = (int *)p;
#ifdef DEBUG_DEVICES		
    printk("wake_up called for queue 0x%x, *p = %d\n", p, *p);
#endif
    if (*ptr > 0) *ptr = *ptr - 1;
}


U64 virttophys(U64 address)
{
    return address;          /* we run in physical mode */
}



int check_disk_change(dev_t dev)
{
	int i;
	struct file_operations * fops;

#ifdef DEBUG_DEVICES		
        printk("check_disk_change called\n");
#endif

	i = MAJOR(dev);
	if (i >= MAX_BLKDEV || (fops = blkdevs[i].fops) == NULL)
		return 0;
	if (fops->check_media_change == NULL)
		return 0;
	if (!fops->check_media_change(dev))
		return 0;

	printk("VFS: Disk change detected on device %d/%d\n",
					MAJOR(dev), MINOR(dev));
	if (fops->revalidate) {
#ifdef DEBUG_DEVICES		
	        printk("...calling revalidate routine @ 0x%p\n", fops->revalidate);
#endif
		fops->revalidate(dev);
	    }
	return 1;

}

int permission(struct inode * inode,int mask)
{
#ifdef DEBUG_DEVICES		
    printk("permission called\n");
#endif
    return 0;
}

void invalidate_buffers(dev_t dev)
{
#ifdef DEBUG_DEVICES		
    printk("invalidate_buffers called\n");
#endif
}

void synch_dev(void)
{
#ifdef DEBUG_DEVICES		
    printk("synch_dev called\n");
#endif
}

void sync_dev(dev_t dev)
{
#ifdef DEBUG_DEVICES		
    printk("sync_dev called\n");
#endif
}

/*
 *  Generic read routine for block devices.  We must set up and call the 
 *  request function for this particular device.
 */
int block_read(struct inode * inode, struct file * filp, char * buf, int count)
{
    int major;

#ifdef DEBUG_DEVICES		
    printk("block_read: called for device 0x%04x, f_pos = %d\n", inode->i_rdev, 
	   filp->f_pos);
#endif

    major = MAJOR(inode->i_rdev);
    if (blk_dev[major].request_fn == NULL) {
	printk("...ERROR: request_fn = NULL\n");
	printk("\tmajor = %d\n", major);
	return 0;
    } else {
	int read;
	int blocksize, block;
	struct buffer_head *bh;
/*
 *  Work out what the blocksize is.
 */
	blocksize = BLOCK_SIZE;
	if (blksize_size[MAJOR(inode->i_rdev)]) {
	    if (blksize_size[MAJOR(inode->i_rdev)][MINOR(inode->i_rdev)])
		blocksize = blksize_size[MAJOR(inode->i_rdev)][MINOR(inode->i_rdev)];
	}
#ifdef DEBUG_DEVICES		
	printk("...block size is %d\n", blocksize);
#endif

/*
 *  We need to read all of the blocks that contain the user's buffer.  We
 *  are at filp->f_pos and the block size is blocksize.
 */
	read = 0;
	while (read < count) {

/*
 *  Which block should we read?
 */
	    block = filp->f_pos / blocksize;
/*
 *  Issue the request to read the file a block at a time.
 */
#ifdef DEBUG_DEVICES		
	    printk("block_read: issuing request to read block %d (size = %d)\n", 
		   block, blocksize);
#endif
	    bh = bread(inode->i_rdev, block, blocksize);
	    if ( !bh ) {
/*
 *  Failed, return just how much we managed to read.
 */
		printk("block_read: read failed (after %d bytes)\n", read);
		return read;          
	    } else {
		int size;
/*
 *  Success, so now update the read count and copy over the data into
 *  the request buffer.  We have just read block "block" and filp->f_pos
 *  tells us where we were in this file before we read this block.  
 *  block * blocksize gives us the offset of the first byte we actually have from 
 *  this file.
 */
		if (read == 0) {
		    int first;
		    int offset;
/*
 *  This is the first block that we've read, (read == 0).
 *  We need only the last part of this block (or all of it).
 */
		    first = block * blocksize;
		    offset = filp->f_pos - first;
		    if (count > blocksize)
			size = blocksize;
		    else
			size = count;
		    memcpy(buf, (char *)(bh->b_data) + offset, size);
		} else {         /* read != 0 */
/*
 *  nth block read.
 *  We need either all of this block or only the first part of it.
 */
       		    if ((count - read) > blocksize) 
			size = blocksize;
		    else
			size = count - read;
		    memcpy(buf, (char *)(bh->b_data), size);
		}
		buf += size;
		read += size;
		filp->f_pos += size;
/*
 *  Have we read enough?
 */
		if (read == count) return read;

	    }
	}              /* end while not read everything */
	return read;
    }                  /* else valid device */
}

int block_write(struct inode * inode, struct file *file, const char *buffer, int a)
{
#ifdef DEBUG_DEVICES		
    printk("block_write called\n");
#endif
    return 1;
}

int block_fsync(struct inode *inode, struct file *file)
{
#ifdef DEBUG_DEVICES		
    printk("block_fsync called\n");
#endif
    return 0;
}

int request_dma(void) 
{
#ifdef DEBUG_DEVICES		
    printk("request_dma called\n");
#endif
    return 0;
}

int __verify_write(unsigned long start, unsigned long size)
{
#ifdef DEBUG_DEVICES		
    printk("__verify_write called\n");
#endif
    return 0;
}

void __wait_on_buffer(struct buffer_head * bh)
{
    unsigned long flags;

    save_flags(flags);
#ifdef DEBUG_DEVICES
    printk("__wait_on_buffer called for bh = 0x%p\n", bh);
#endif
    sti();
    while (bh->b_lock) __asm__ __volatile__("": : :"memory");
     restore_flags(flags);
}



int verify_area(int type, const void *where, unsigned long size)
{
#ifdef DEBUG_DEVICES		
    printk("verify_area: called\n");
#endif
    return 1;
}

static struct buffer_head *bh = NULL;
static char *buf = NULL;
struct buffer_head * getblk(dev_t dev, int block, int size)
{
#ifdef DEBUG_DEVICES
    printk("getblk: called (dev = 0x%x), size = %d\n", dev, size);
#endif

    /* allocate some memory for it (if we have to). */
    if (bh == NULL) {
	bh = (struct buffer_head *)malloc(sizeof(struct buffer_head));
	/* make sure that the buffer is size aligned */
	buf = malloc(size * 2) + size;
	buf = (char *)((unsigned long)buf & ~(size - 1));
#ifdef DEBUG_DEVICES
	printk("getblk: allocating *new* bh (@ 0x%x) and buf (@ 0x%x)\n",
	       bh, buf);
#endif
    }
    memset((void *)bh, 0, sizeof(struct buffer_head));
    memset((void *)buf, 0, size);

    /* now fill it out */
    bh->b_data = buf;
    bh->b_req=0;
    bh->b_dev = dev;
    bh->b_blocknr = block;
    bh->b_count = 1;
    bh->b_dirt = 0;
    bh->b_lock = 0;
    bh->b_uptodate = 0;
    bh->b_flushtime = 0;
    bh->b_prev = NULL;
    bh->b_next = NULL;
    bh->b_prev_free = NULL;
    bh->b_next_free = NULL;
    bh->b_this_page = NULL;
    bh->b_reqnext = NULL;
    bh->b_size = size;

    return bh;
}

void brelse(struct buffer_head * buf)
{
#ifdef DEBUG_DEVICES		
    printk("brelse: called\n");
#endif

    if (buf->b_count) buf->b_count--;
}

int register_blkdev(unsigned int major, const char * name, struct file_operations *fops)
{
#ifdef DEBUG_DEVICES		
        printk("register_blkdev: called for %s\n", name);
#endif

	if (major >= MAX_BLKDEV)
		return -EINVAL;
	if (blkdevs[major].fops)
		return -EBUSY;
	blkdevs[major].name = name;
	blkdevs[major].fops = fops;
	return 0;
}

int unregister_blkdev(unsigned int major, const char * name)
{
#ifdef DEBUG_DEVICES		
        printk("unregister_blkdev: called for %s\n", name);
#endif

	if (major >= MAX_BLKDEV)
		return -EINVAL;
	if (!blkdevs[major].fops)
		return -EINVAL;
	if (strcmp(blkdevs[major].name, name))
		return -EINVAL;
	blkdevs[major].name = NULL;
	blkdevs[major].fops = NULL;
	return 0;
}

void free_dma(unsigned int dmanr)
{
#ifdef DEBUG_DEVICES		
    printk("free_dma: called\n");
#endif
}


/* ------------------------------------------------------------ */

asmlinkage void schedule(void)
{

#ifdef DEBUG_DEVICES		
    printk("schedule(): called\n");
#endif

#if 0
    /* drop the ipl for a while so that interrupts can happen */
    do {
	unsigned long flags;

	save_flags(flags);
	sti();
	udelay(100);
	restore_flags(flags);
    } while (0);
#endif
}

void refile_buffer(struct buffer_head * buf)
{
#ifdef DEBUG_DEVICES		
    printk("refile_buffer(): called with buf  = 0x%p\n", buf);
#endif
}

void __down(struct semaphore * sem)
{
#ifdef DEBUG_DEVICES		
    printk("__down(): called\n");
#endif

#if 0
    /* drop the ipl for a while so that interrupts can happen */
    do {
	unsigned long flags;

	save_flags(flags);
	sti();
	udelay(100);
	restore_flags(flags);
    } while (0);
#endif
}

/*
 * bread() reads a specified block and returns the buffer that contains
 * it. It returns NULL if the block was unreadable.
 */
struct buffer_head * bread(dev_t dev, int block, int size)
{
	struct buffer_head * bh;

#ifdef DEBUG_DEVICES		
	printk("bread: called\n");
#endif

	if (!(bh = getblk(dev, block, size))) {
		printk("VFS: bread: READ error on device %d/%d\n",
						MAJOR(dev), MINOR(dev));
		return NULL;
	}
	if (bh->b_uptodate)
		return bh;
	ll_rw_block(READ, 1, &bh);
	wait_on_buffer(bh);
	if (bh->b_uptodate)
		return bh;
	brelse(bh);
	return NULL;
}

struct buffer_head * breada(dev_t dev, int block, int bufsize,
	unsigned int pos, unsigned int filesize)
{
#ifdef DEBUG_DEVICES		
    printk("breada(): called\n");
#endif
    return 0;
}

void wait_for_keypress(void)
{
    uart_getchar(COM1);
}

/*
 * Called every time a block special file is opened
 */
int blkdev_open(struct inode * inode, struct file * filp)
{
	int i;

#ifdef DEBUG_DEVICES		
	printk("blkdev_open: called\n");
#endif

	i = MAJOR(inode->i_rdev);
	if (i >= MAX_BLKDEV || !blkdevs[i].fops)
		return -ENODEV;
	filp->f_op = blkdevs[i].fops;
	if (filp->f_op->open)
		return filp->f_op->open(inode,filp);
	return 0;
}	

void kfree (void *ptr,int size)
{
#ifdef DEBUG_DEVICES		
    printk("kfree(): called\n");
#endif
}

void * kmalloc (size_t size, int priority)
{
    void *ptr;

    ptr = (void *)malloc(size);
#ifdef DEBUG_DEVICES		
    printk("kmalloc(): called, returning 0x%p\n", ptr);
#endif
    return ptr;
}

void invalidate_inodes(dev_t dev)
{
#ifdef DEBUG_DEVICES		
    printk("invalidate_inodes(): called\n");
#endif
}

int fsync_dev(dev_t dev)
{
#ifdef DEBUG_DEVICES		
    printk("fsync_dev(): called\n");
#endif
    return 0;
}

int rd_init(long mem_start, int length)
{
#ifdef DEBUG_DEVICES		
    printk("rd_init: called (why?)\n");
#endif
    return (length);
}

void rd_load(void)
{
#ifdef DEBUG_DEVICES		
    printk("rd_load: called (why?)\n");
#endif
}

unsigned long __get_free_pages(int priority, unsigned long order, unsigned long limit)
{
    int size;
    char *ptr;
    int count, pfn;

#ifdef DEBUG_DEVICES
    printk("__get_free_pages: called, priority = %d, order = %d, limit = %d\n",
	   priority, order, limit);
#endif

    /* the size of memory that we want is based on order */
    size = PAGE_SIZE << order;
    count = size / PAGE_SIZE;
    pfn = allocate_many_pfn(count, TEMPORARY_PAGE);
    ptr = _PFN_TO_PHYSICAL(pfn);
    
#ifdef DEBUG_DEVICES
    printk("__get_free_pages(): returning 0x%p\n", ptr);
#endif
    return (unsigned long)ptr;
}

void free_pages(unsigned long addr, unsigned long order)
{
#ifdef DEBUG_DEVICES
    printk("free_pages: called\n");
#endif
}

void nop(void)
{
#ifdef DEBUG_DEVICES		
    printk("nop: called\n");
#endif
}

#endif
/*
 * entUna has a different register layout to be reasonably simple. It
 * needs access to all the integer registers (the kernel doesn't use
 * fp-regs), and it needs to have them in order for simpler access.
 *
 * Due to the non-standard register layout (and because we don't want
 * to handle floating-point regs), we disallow user-mode unaligned
 * accesses (we'd need to do "verify_area()" checking, as well as
 * do a full "ret_from_sys_call" return).
 *
 * Oh, btw, we don't handle the "gp" register correctly, but if we fault
 * on a gp-register unaligned load/store, something is _very_ wrong
 * in the kernel anyway..
 */
struct allregs {
	unsigned long regs[32];
	unsigned long ps, pc, gp, a0, a1, a2;
};

asmlinkage void do_entUna(void * va, unsigned long opcode, unsigned long reg,
	unsigned long a3, unsigned long a4, unsigned long a5,
	struct allregs regs)
{
	static int cnt = 0;

	if (regs.ps & 8)
	        panic("do_entUna: failed(1) \n");
	if (++cnt < 5)
		printk("Unaligned trap at %016lx: %p %lx %ld\n",
			regs.pc, va, opcode, reg);
	/* $16-$18 are PAL-saved, and are offset by 19 entries */
	if (reg >= 16 && reg <= 18)
		reg += 19;
	switch (opcode) {
		case 0x28: /* ldl */
			*(reg+regs.regs) = (int) ldl_u(va);
			return;
		case 0x29: /* ldq */
			*(reg+regs.regs) = ldq_u(va);
			return;
		case 0x2c: /* stl */
			stl_u(*(reg+regs.regs), va);
			return;
		case 0x2d: /* stq */
			stq_u(*(reg+regs.regs), va);
			return;
	}
	printk("Bad unaligned kernel access at %016lx: %p %lx %ld\n",
		regs.pc, va, opcode, reg);
	panic("do_entUna: failed(2) \n");
}


