/*
 * paradis@amt.tay1.dec.com
 */

#include "miniboot.h"
#include "uart.h"

extern volatile unsigned long jiffies;

#ifdef MINI_SERIAL_IO

void kbd_init(void)
{
}

int kbd_getc_with_timeout(int nsecs)
{
    unsigned long endjifs;

    if(nsecs) {
        endjifs = jiffies + ((unsigned long)nsecs * 1000L);
    }

    while (1) {
	/* Input from both uart */
	if(uart_charav(COM1)) return(uart_getchar(COM1));

	if(nsecs) {
	    if(jiffies > endjifs) {
		return -1;
	    }
	}
    }
}

#else
/* This is a *very* simple polled keyboard driver.  The scancode
 * processing is primitive, and it only understands CTRL and SHIFT
 * (and not both at once, either!  CTRL overrides SHIFT...)
 */

/* These are the only "special" keyboard functions we'll support */
#define KB_CTRL		0xf0
#define KB_SHIFT	0xf1
#define KB_CAPSLOCK	0xf2

/* These are just placeholders... we don't assign them any real
 * value or function...
 */
#define KB_NUMLOCK	0
#define KB_SCRLOCK	0
#define KB_ALT		0

/* This is the scancode mapping table for "plain" characters
 * (no ctrl or shift)
 */
static unsigned char plain_charmap[] = {
    0, '\033', '1', '2', '3', '4', '5', '6',		/* 0x00 */
    '7', '8', '9', '0', '-', '=', '\010', '\011',	/* 0x08 */
    'q', 'w', 'e', 'r', 't', 'y', 'u', 'i',		/* 0x10 */
    'o', 'p', '[', ']', '\015', KB_CTRL, 'a', 's',      /* 0x18 */
    'd', 'f', 'g', 'h', 'j', 'k', 'l', ';',             /* 0x20 */
    '\047', '`', KB_SHIFT, '\\', 'z', 'x', 'c', 'v',    /* 0x28 */
    'b', 'n', 'm', ',', '.', '/', KB_SHIFT, '*',        /* 0x30 */
    KB_ALT, '\040', KB_CAPSLOCK, 0, 0, 0, 0, 0,		/* 0x38 */
    0, 0, 0, 0, 0, KB_NUMLOCK, KB_SCRLOCK, '7',		/* 0x40 */
    '8', '9', '-', '4', '5', '6', '+', '1',		/* 0x48 */
    '2', '3', '0', '.', 0, 0, 0, 0,			/* 0x50 */
    0, 0, 0, 0, 0, 0, 0, 0,				/* 0x58 */
    0, 0, 0, 0, 0, 0, 0, 0,				/* 0x60 */
    0, 0, 0, 0, 0, 0, 0, 0,				/* 0x68 */
    0, 0, 0, 0, 0, 0, 0, 0,				/* 0x70 */
    0, 0, 0, 0, 0, 0, 0, 0				/* 0x78 */
};


/* This is the scancode mapping table for shifted characters */
static unsigned char shift_charmap[] = {
    0, '\033', '!', '@', '#', '$', '%', '^',		/* 0x00 */
    '&', '*', '(', ')', '_', '+', '\010', '\011',	/* 0x08 */
    'Q', 'W', 'E', 'R', 'T', 'Y', 'U', 'I',		/* 0x10 */
    'O', 'P', '{', '}', '\015', KB_CTRL, 'A', 'S',	/* 0x18 */
    'D', 'F', 'G', 'H', 'J', 'K', 'L', ':',		/* 0x20 */
    '"', '~', KB_SHIFT, '|', 'Z', 'X', 'C', 'V',	/* 0x28 */
    'B', 'N', 'M', '<', '>', '?', KB_SHIFT, '*',	/* 0x30 */
    KB_ALT, '\040', KB_CAPSLOCK, 0, 0, 0, 0, 0,		/* 0x38 */
    0, 0, 0, 0, 0, KB_NUMLOCK, KB_SCRLOCK, '7',		/* 0x40 */
    '8', '9', '-', '4', '5', '6', '+', '1',		/* 0x48 */
    '2', '3', '0', '.', 0, 0, 0, 0,			/* 0x50 */
    0, 0, 0, 0, 0, 0, 0, 0,				/* 0x58 */
    0, 0, 0, 0, 0, 0, 0, 0,				/* 0x60 */
    0, 0, 0, 0, 0, 0, 0, 0,				/* 0x68 */
    0, 0, 0, 0, 0, 0, 0, 0,				/* 0x70 */
    0, 0, 0, 0, 0, 0, 0, 0				/* 0x78 */
};

/* This is the scancode mapping table for control characters */
static unsigned char ctrl_charmap[] = {
    '\000', '\033', '1', '\000', '\033', '\034', '\035', '\036',     /* 0x00 */
    '\037', '\177', '9', '0', '-', '=', '\030', '\011',              /* 0x08 */
    '\021', '\027', '\005', '\022', '\024', '\031', '\025', '\011',  /* 0x10 */
    '\017', '\020', '\033', '\035', '\015', KB_CTRL, '\001', '\023', /* 0x18 */
    '\004', '\006', '\007', '\010', '\012', '\013', '\014', ';',     /* 0x20 */
    '\047', '\036', KB_SHIFT, '\034', '\032', '\030', '\003', '\026',/* 0x28 */
    '\002', '\016', '\015', ',', '.', '/', KB_SHIFT, '*',            /* 0x30 */
    KB_ALT, '\000', KB_CAPSLOCK, 0, 0, 0, 0, 0,                      /* 0x38 */
    0, 0, 0, 0, 0, KB_NUMLOCK, KB_SCRLOCK, '7',                      /* 0x40 */
    '8', '9', '-', '4', '5', '6', '+', '1',                          /* 0x48 */
    '2', '3', '0', '.', '\000', '\000', '\000', 0,                   /* 0x50 */
    0, '\000', '\000', '\000', '\000', '\000', '\000', '\000',       /* 0x58 */
    '\000', '\000', '\000', '\000', '\000', '\000', '\000', '\000',  /* 0x60 */
    '\000', '\000', '\000', '\000', '\000', '\000', '\000', '\000',  /* 0x68 */
    '\000', '\000', '\000', '\000', '\000', '\000', '\000', '\000',  /* 0x70 */
    '\000', '\000', '\000', '\000', '\000', '\000', '\000', '\000'   /* 0x78 */
};

/* These are the keyboard states we can support */
enum kbdstates {
    KBD_NORMAL,
    KBD_SHIFTED,
    KBD_CTRL,
};

enum kbdstates	keyboard_state = KBD_NORMAL;

#define SCANCODE_KEYUP	0x80
#define SCANCODE_LEFTSHIFT	0x2a
#define SCANCODE_RIGHTSHIFT	0x36
#define	SCANCODE_CTRL		0x1d
#define SCANCODE_CAPSLOCK	0x3a
#define SCANCODE_EXTENDED	0xe0


static inline void kbd_wait(void)
{
    int	i;

    /* Wait for the keyboard controller to fetch its command... */
    for(i = 0; i < 0x100000; i++) {
	if ((inb_p(0x64) & 0x02) == 0) {
		return;
	}
    }
}

void kbd_init(void)
{
    /* Set up keyboard in PC/AT, noninterrupt mode */
    kbd_wait();
    outb(0x60, 0x64);	/* Command - write PS/2 mode register */
    kbd_wait();
#if 1
    outb(0x65,0x60);	/* KCC | DMS | SYS | EKI */
#else
    outb(0x60, 0x60);	/* KCC | DMS */
#endif
    kbd_wait();
}

int kbd_getc_with_timeout(int nsecs)
{
    int		scancode;
    int		character;
    unsigned long endjifs;

    if(nsecs) {
        endjifs = jiffies + ((unsigned long)nsecs * 1000L);
    }

    for(;;) {

	for(;;) {
	    /* Read a scancode from the keyboard... */
	    if ((inb_p(0x64) & 0x21) == 1) {
		scancode = inb(0x60);
		break;
	    }
#ifdef MINI_SERIAL_ECHO
	    /* Allow input from both uart & kbd */
	    if(uart_charav(COM1)) {
		return(uart_getchar(COM1));
	    }
#endif

	    if(nsecs) {
		if(jiffies > endjifs) {
		    return -1;
		}
	    }
	}

	/* If it's an extended function, ignore it... */
	if(scancode == SCANCODE_EXTENDED) {
	    continue;
	}

	/* Process certain key releases first...  */
	if(scancode & SCANCODE_KEYUP) {
	    scancode &= ~SCANCODE_KEYUP;

	    /* Did we release a SHIFT key? */
	    if(((scancode == SCANCODE_LEFTSHIFT) || 
		(scancode == SCANCODE_RIGHTSHIFT)) &&
	       (keyboard_state == KBD_SHIFTED)) {
		keyboard_state = KBD_NORMAL;
		continue;
	    }

	    /* Did we release a CTRL key? */
	    if((scancode == SCANCODE_CTRL) && (keyboard_state == KBD_CTRL)) {
		keyboard_state = KBD_NORMAL;
		continue;
	    }

	    /* Other releases aren't important to us... */
	    continue;
	}

	/* It's a keypress.  Is it a ctrl? */ 
	if(scancode == SCANCODE_CTRL) {
	    keyboard_state = KBD_CTRL;
	    continue;
	}

	/* Is it a shift? */
	if(((scancode == SCANCODE_LEFTSHIFT) || 
	    (scancode == SCANCODE_RIGHTSHIFT)) &&
	   (keyboard_state != KBD_CTRL)) {
	    keyboard_state = KBD_SHIFTED;
	    continue;
	}

	/* Is it a capslock? */
	if((scancode == SCANCODE_CAPSLOCK) && (keyboard_state != KBD_CTRL)) {
	    if(keyboard_state == KBD_NORMAL) {
		keyboard_state = KBD_SHIFTED;
	    }
	    else {
		keyboard_state = KBD_NORMAL;
	    }
	    continue;
	}

	/* It's none of these.  Find the right table and translate... */
	switch(keyboard_state) {
	    case KBD_NORMAL:
		character = plain_charmap[scancode];
		break;
	    case KBD_SHIFTED:
		character = shift_charmap[scancode];
		break;
	    case KBD_CTRL:
		character = ctrl_charmap[scancode];
		break;
	}

	if(character) {
	    return(character);
	}
    }
}

#endif

/* Wait for keyboard, and return whatever data the it keyboard has for us... */
int kbd_getc(void)
{
    return(kbd_getc_with_timeout(0));
}



/* Very simple-minded get-string function.  Understands rubout and CR, 
 * and character echo, but not much else...  Cuts off any trailing
 * newline, and null-terminates the return string.  Returns the number
 * of characters read.
 */
int kbd_gets(char * charbuf, int maxlen)
{
    int		bufpos;
    char	c;

    bufpos = 0;
    for(;;) {
	c = kbd_getc();
	switch(c) {
	    case 0x08:
	    case 0x7f:
		/* Backspace or rubout means to get rid of the
		 * previous character, if any.
		 */
		if(bufpos > 0) {
		    bufpos--;
		    vga_putchar('\b');
		    vga_putchar(' ');
		    vga_putchar('\b');
#ifdef MINI_SERIAL_ECHO
		    uart_putchar(COM1, '\b');
		    uart_putchar(COM1, ' ');
		    uart_putchar(COM1, '\b');
#endif
		    
		}
		break;

	    case 0x0d:
	    case 0x0a:
		/* CR or LF means to null-terminate the string
		 * and return the number of characters read.
		 */
#ifdef MINI_SERIAL_ECHO
		vga_putchar('\r');
		vga_putchar('\n');
#else
		uart_putchar(COM1, '\r');
		uart_putchar(COM1, '\n');
#endif
		charbuf[bufpos] = '\0';
		return(bufpos);

	    default:
		/* Stuff this character into the buffer (if we have room) 
		 * and advance the character pointer...
		 */
		if((bufpos+1) < maxlen) {
		    charbuf[bufpos++] = c;
		    vga_putchar(c);
#ifdef MINI_SERIAL_ECHO
		    uart_putchar(COM1, c);
#endif
		}
		break;
	}
    }
}
