#include <alpha/regdef.h>
#include <alpha/pal.h>
#include <asm/system.h>
#include <linux/config.h>

#ifdef CALL_PAL_BROKEN
#define call_pal	.long
#endif

/* Here are some VMS PALcode functions that we need in order to
 * switch to OSF PALcode...
 */

#define PAL_mfpr_pcbb    0x12
#define PAL_ldqp	 0x03
#define PAL_stqp	 0x04
#define PAL_halt_entry	 0x00
#define PAL_cserve_entry 0x09
#define SWPPAL		 10

	.set	noreorder

#ifdef __linux__
	.set singlegp
#endif

	.text

	.globl	__start
	.ent	__start
__start:
#ifdef HALT_ON_STARTUP
	call_pal	PAL_halt	# offset 0
#else
	br	1f
	bis	zero,zero,zero		# spacer
	/* 
	 *  Make room for arguments to be passed to the Linux Kernel.
	 *
 	 *  offset 	meaning
	 *
	 *    0*8       Linux kernel entry point
	 *    1*8       First free pfn
	 *    2*8       Last free pfn
	 *    3*8       last pfn
	 *    4*8       ramdisk address 
	 *    5*8       ramdisk pages (each 1024 bytes long)
	 *    6*8       boot_flags
	 */
	.quad	0,0,0,0,0,0,0           
#endif
/*
 *  Get the gp register set up correctly.
 */
1:	bis	zero,zero,zero		# spacer
	br	t0,2f			# offset 4
	.quad	_gp			# offset 8
2:	ldq	gp,0(t0)		# offset 16
/*
 *  make sure that we're in kseg.
 */
	jsr	ra, __turn_on_kseg

/*
 *  zero out data at the end of the image
 */
	jsr	ra, doinitdata
/*
 *  Zero a few arguments and jump into boot_main.
 */
	bis	zero,zero,a0
	bis	zero,zero,a1
	bis	zero,zero,a2

	jsr	ra, boot_main

	call_pal PAL_halt
	.end	__start

/*
 *  Entry point for when we've jumped to the new palcode inside the
 *  miniloader.
 */
	.globl	__start_again
	.ent	__start_again
__start_again:
	br	t0,3f			# offset 4
	.quad	_gp			# offset 8
3:	ldq	gp,0(t0)		# offset 16
/*
 *  Zero a few arguments and jump into boot_main_again().
 */
	bis	zero,zero,a0
	bis	zero,zero,a1
	bis	zero,zero,a2

	jsr	ra, boot_main_cont

	call_pal PAL_halt
	.end 	__start_again

#if 0
/* 
 * Zeros out one page of physical memory.  A0 = pfn of physical page
 * to zero out.  Uses physical mapping.
 */
	.globl	zeropage_phys
	.ent	zeropage_phys
zeropage_phys:
	sll	a0,13,a0		# Convert r16 from pfn to physaddr
	lda	t0,1024(zero)		# Number of quads in a page
$301:	stq	zero,0(a0)		# stuff a quadword
	lda	a0,8(a0)		# Increment pointer
	subq	t0,1,t0			# Decrement count
	bne	t0,$301			# Loop if not zero
	ret	zero,(ra)		# done!
	.end	zeropage_phys
#endif

	.globl	imb
	.ent	imb
imb:
	call_pal	PAL_imb
	ret	zero,(ra)
	.end	imb

	.globl	entMM
	.ent	entMM
entMM:	call_pal	PAL_halt
	.end	entMM

	.globl	entIF
	.ent	entIF
entIF:	subq	a0,3,a0		# If a0 == 3, this is an FEN fault.  
	beq	a0,1f		# Try setting FEN to 1 and trying again...
	call_pal	PAL_halt
1:	lda	sp,-48(sp)
	stq	a0,0(sp)
	stq	t0,8(sp)
	stq	t8,16(sp)
	stq	t9,24(sp)
	stq	t10,32(sp)
	stq	t11,40(sp)
	lda	a0,1(zero)
	call_pal	PAL_wrfen
	ldq	a0,0(sp)
	ldq	t0,8(sp)
	ldq	t8,16(sp)
	ldq	t9,24(sp)
	ldq	t10,32(sp)
	ldq	t11,40(sp)
	lda	sp,48(sp)
	call_pal	PAL_rti

	.end	entIF



.align 3
.globl	entUna
.ent	entUna
entUna:
	lda $30,-256($30)
	stq $0,0($30)
	stq $1,8($30)
	stq $2,16($30)
	stq $3,24($30)
	stq $4,32($30)
	stq $5,40($30)
	stq $6,48($30)
	stq $7,56($30)
	stq $8,64($30)
	stq $9,72($30)
	stq $10,80($30)
	stq $11,88($30)
	stq $12,96($30)
	stq $13,104($30)
	stq $14,112($30)
	stq $15,120($30)
	/* 16-18 PAL-saved */
	stq $19,152($30)
	stq $20,160($30)
	stq $21,168($30)
	stq $22,176($30)
	stq $23,184($30)
	stq $24,192($30)
	stq $25,200($30)
	stq $26,208($30)
	stq $27,216($30)
	stq $28,224($30)
	stq $29,232($30)
	stq $30,240($30)
	stq $31,248($30)
	bis	zero,zero,zero		# spacer
	br	t0,1f			# offset 4
	.quad	_gp			# offset 8
1:	ldq	gp,0(t0)		# offset 16
	lda $27,do_entUna
	jsr $26,($27),do_entUna
	ldq $0,0($30)
	ldq $1,8($30)
	ldq $2,16($30)
	ldq $3,24($30)
	ldq $4,32($30)
	ldq $5,40($30)
	ldq $6,48($30)
	ldq $7,56($30)
	ldq $8,64($30)
	ldq $9,72($30)
	ldq $10,80($30)
	ldq $11,88($30)
	ldq $12,96($30)
	ldq $13,104($30)
	ldq $14,112($30)
	ldq $15,120($30)
	/* 16-18 PAL-saved */
	ldq $19,152($30)
	ldq $20,160($30)
	ldq $21,168($30)
	ldq $22,176($30)
	ldq $23,184($30)
	ldq $24,192($30)
	ldq $25,200($30)
	ldq $26,208($30)
	ldq $27,216($30)
	ldq $28,224($30)
	ldq $29,232($30)
	ldq $30,240($30)
	lda $30,256($30)
	call_pal	PAL_rti
.end entUna

/*
 * These offsets must match with "struct hae" in io.h:	
 */
#define HAE_CACHE	0
#define HAE_REG		8

/*
 * This defines the normal kernel pt-regs layout.
 *
 * regs 9-15 preserved by C code
 * regs 16-18 saved by PAL-code
 * regs 29-30 saved and set up by PAL-code
 */
#define SAVE_ALL			\
	subq	$30,160,$30;		\
	stq	$0,0($30);		\
	stq	$1,8($30);		\
	stq	$2,16($30);		\
	stq	$3,24($30);		\
	stq	$4,32($30);		\
	stq	$5,40($30);		\
	stq	$6,48($30);		\
	stq	$7,56($30);		\
	stq	$8,64($30);		\
	stq	$19,72($30);		\
	stq	$20,80($30);		\
	stq	$21,88($30);		\
	stq	$22,96($30);		\
	stq	$23,104($30);		\
	stq	$24,112($30);		\
	stq	$25,120($30);		\
	stq	$26,128($30);		\
	stq	$27,136($30);		\
	stq	$28,144($30);		\
	lda	$2,hae;			\
	ldq	$2,HAE_CACHE($2);	\
	stq	$2,152($30)

#define RESTORE_ALL			\
	lda	$8,hae;			\
	ldq	$7,HAE_CACHE($8);	\
	ldq	$6,152($30);		\
	subq	$7,$6,$5;		\
	beq	$5,99f;			\
	ldq	$7,HAE_REG($8);		\
	stq	$6,HAE_CACHE($8);	\
	stq	$6,0($7);		\
	mb;				\
99:;					\
	ldq	$0,0($30);		\
	ldq	$1,8($30);		\
	ldq	$2,16($30);		\
	ldq	$3,24($30);		\
	ldq	$4,32($30);		\
	ldq	$5,40($30);		\
	ldq	$6,48($30);		\
	ldq	$7,56($30);		\
	ldq	$8,64($30);		\
	ldq	$19,72($30);		\
	ldq	$20,80($30);		\
	ldq	$21,88($30);		\
	ldq	$22,96($30);		\
	ldq	$23,104($30);		\
	ldq	$24,112($30);		\
	ldq	$25,120($30);		\
	ldq	$26,128($30);		\
	ldq	$27,136($30);		\
	ldq	$28,144($30);		\
	addq	$30,160,$30

	.text

	.align 3
	.globl	entInt
	.ent	entInt
entInt:
	/* set up the gp */
	br gp,entInt2
entInt2:
	ldgp gp,0(gp)

	SAVE_ALL

	lda	$27,do_entInt
	jsr	$26,($27),do_entInt

	RESTORE_ALL

	call_pal	PAL_rti
	.end entInt

	.align 3
	.globl	wrent
	.ent	wrent
wrent:
	call_pal	PAL_wrent
	ret ($26)
	.end wrent

	.globl	panic
	.ent	panic 2
panic:
	.option O1
	.frame sp, 0, ra

	bis a4, zero, a5
	bis a3, zero, a4
	bis a2, zero, a3
	bis a1, zero, a2
	bis a0, zero, a1
	bis $26, zero, a0
	jsr zero, miniboot_panic
	.end 	panic

#if 1
	.align 3
	.globl	rdmces
	.ent	rdmces
rdmces:
	call_pal PAL_rdmces
	ret ($26)
	.end rdmces

	.align 3
	.globl	wrmces
	.ent	wrmces
wrmces:
	call_pal PAL_wrmces
	ret ($26)
	.end wrmces


#endif

/*
 * Where compilers cannot dereference quadword addresses
 * (ie NT), then we must have a assembler routines for this.
 */
	.globl	ReadQ
	.ent	ReadQ 2
ReadQ:
	.option	O1
	.frame	sp, 0, ra

	ldq	v0, 0(a0)
	ret	zero, (ra)
	.end	ReadQ

	.globl	WriteQ
	.ent	WriteQ 2
WriteQ:
	.option	O1
	.frame	sp, 0, ra

	stq	a1, 0(a0)
	ret	zero, (ra)
	.end	WriteQ

	.globl	halt
	.ent	halt 2
halt:
	.option	O1
	.frame	sp, 0, ra
	call_pal PAL_halt_entry
	ret	zero, (ra)
	.end	halt


	.globl  rdsp
	.ent	rdsp
rdsp:
	.option O1
	.frame sp, 0, ra
	bis sp, zero, v0
	ret	zero, (ra)
	.end rdsp

	.globl  wrsp
	.ent	wrsp
wrsp:
	.option O1
	.frame sp, 0, ra
	bis a0, zero, sp
	ret	zero, (ra)
	.end wrsp

	.globl	cServe
	.ent	cServe 2
cServe:
	.option	O1
	.frame	sp, 0, ra
	call_pal PAL_cserve_entry
	ret	zero, (ra)
	.end	cServe


/*
 * read_cpureg - Read a CPU register
 */

	.globl	read_cpureg_long
	.ent	read_cpureg_long
read_cpureg_long:
	ldl	v0,0(a0)
	ret	zero,(ra)
	.end	read_cpureg_long

/*
 * write_cpureg_long - Write a CPU register
 */
	.globl	write_cpureg_long
	.ent	write_cpureg_long
write_cpureg_long:
	stl	a0,0(a1)
	mb
	ret	zero,(ra)
	.end	write_cpureg_long

/*
 * read_cpureg_long - Read a CPU register
 */

	.globl	read_cpureg_quad
	.ent	read_cpureg_quad
read_cpureg_quad:
	ldq	v0,0(a0)
	ret	zero,(ra)
	.end	read_cpureg_quad

/*
 * write_cpureg_quad - Write a CPU register
 */
	.globl	write_cpureg_quad
	.ent	write_cpureg_quad
write_cpureg_quad:
	stq	a0,0(a1)
	mb
	ret	zero,(ra)
	.end	write_cpureg_quad

/* Set the current interrupt priority level, returning the previous IPL 
 *
 * Calling sequence:
 *
 *	old_ipl = ipl(new_ipl)
 *
 * The SWPIPL pal function does this all for us 8-)
 */

	.globl	ipl
	.ent	ipl
ipl:
	call_pal	PAL_swpipl
	ret		zero,(ra)
	.end	ipl


	.globl	wrfen
	.ent	wrfen
wrfen:
	lda a0, 1(zero)
	call_pal PAL_wrfen
	ret  	zero, (ra)
	.end wrfen


	.globl	swppal
	.ent	swppal
swppal: call_pal PAL_swppal
	.end	swppal

/*
 *  Convert to running in OSF KSEG (directly mapped address space).
 */
	.globl 	__turn_on_kseg
	.ent	__turn_on_kseg
__turn_on_kseg:
/*
 * First, the return address (which wasn't KSEG).
 */
	lda	t1, -32768(zero)
	sll	t1, 27, t1
	bis	ra, t1, ra
/*
 * Now the stack pointer.
 */
	bis 	sp, t1, sp

	ret	zero,(ra)
	.end	__turn_on_kseg

#define PAL_NT_di        0x08
#define PAL_NT_rdpsr     0x1a
#define PAL_NT_imb       0x86
#define PAL_NT_swpirql   0x07

	.align 7
	.globl	__arcentry
	.ent	__arcentry
__arcentry:
#if 1
	lda	a0, 7(zero)
	call_pal PAL_NT_swpirql
	bis	zero,zero,zero		# spacer
#else
	bis	zero,zero,zero		# spacer
#endif
	br	t0, 3f
	/* pcb */
	.quad	0x00200000    /* ksp              	*/
	.quad	0             /* usp              	*/
	.quad	0             /* ptbr             	*/
	.quad	0             /* asn and pcc      	*/
	.quad	0x00000001    /* unique and fen   	*/
	.quad	0             /* reserved         	*/

	/* swppal arguments */
	.quad	0x200000      /* OSF PALbase      	*/
	.quad   0x210000      /* Start PC         	*/
	.quad   0x0           /* PCB address      	*/
	.quad   0x1           /* vptbr            	*/

	/* ABOX and ICCSR initial values 	  	*/
	.quad   0x000006f8    /* iccsr            	*/
	.quad	0x00000428    /* abox             	*/
3:
	/* t0 now contains the address of the first quadword. */
	ldq	a1, 6*8(t0)   /* New palbase      	*/
	ldq     a2, 7*8(t0)   /* New console PC   	*/
	bis	t0, zero, a3  /* PCB address      	*/
	sll	a3, 33, a3
	srl 	a3, 33, a3
	ldq     a4, 9*8(t0)   /* vptbr value      	*/

	ldq	t7, 10*8(t0)  /* iccsr  	  	*/
	ldq	t8, 11*8(t0)  /* abox   	  	*/

	/* make the swppal call (to WNT PALcode). */
	bis	zero,zero,zero		# spacer
	br	a0, 4f

	/* ersatz palcode. */
	/* get the arguments into the right order.             */
	bis 	a1, zero, a0
	bis 	a2, zero, a1
	bis 	a3, zero, a2
	bis	a4, zero, a3

	/* write the iccsr and abox registers                  */
	sll	t7, 32, t7
	.long	0x750800A2    /* mtpr t7, pt2_iccsr            */
	.long	0x76D6004E    /* mtpr t8, aboxCtl              */
 
	/* now branch to the real OSF PALcode                  */
	jmp	zero,(a0),0

4:	
	/* a0 now contains the address of my ersatz palcode.   */
	call_pal PAL_swppal
	ret	zero,(ra)
	.end	__arcentry


	.align 7
	.globl	__relocate
	.ent	__relocate
__relocate:
	bis	zero,zero,zero		# spacer
	br	t0, 5f
	.quad	PALCODE_AT;  		# final address
	.quad	0x200000;		# current address
	.quad	0x008000;               # size of image (in quadwords)
	.quad	LOADER_AT; 		# entry point of image.
5:					# t0 now contains address of
					# data block
	ldq	t1, 0*8(t0)	
	ldq	t2, 1*8(t0)
	ldq	t3, 2*8(t0)
	ldq	t4, 3*8(t0)
	ldq	t5, 3*8(t0)

	bis	t1, zero, sp		# set up the stack pointer
	bis	zero, 7, a0		# bump the ipl
	call_pal	PAL_swpipl

6:
	ldq	t4, 0(t2)		# get next data quadword
	stq	t4, 0(t1)		# save it
	lda	t2, 8(t2)		# increment from pointer
	lda	t1, 8(t1)		# increment to pointer
	subq	t3, 1, t3		# decrement the quadword count
	bgt	t3, 6b

	jsr	zero, (t5)		# and jump to it.

	.end	__relocate
