/*
 *  This module contains miniboot specific definitions and macros.
 */

#ifndef MINIBOOT_H
#define MINIBOOT_H 1

#include <asm/io.h>

#ifndef U64 
# define U64 unsigned long long
#endif

#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif

/********************************************************************************
 * external routine declarations.                                               *
 ********************************************************************************/

extern void WriteQ(U64 where, U64 value);
extern U64  ReadQ(U64 where);

#ifndef LINUX_AT
extern int  ext2_mount(long device);
extern void ext2_close(int fd);
#endif

extern char uart_getchar(int port);
extern void  uart_putchar(int port, char c);
extern int  uart_init(void);
extern int uart_charav(int port);

extern int ipl(int newipl);
extern void wrfen(void);
extern void zeropage_phys(int pfn);
extern U64 cServe(U64 a1, U64 a2, U64 a3);
extern void wrsp(U64 sp);

extern void cpu_init(void);
extern void udelay(unsigned long usecs);

extern void device_init(void);

extern void init_IRQ(void);

extern void init_timers(void);
extern void timer_interrupt(void);

extern void vga_init(void);
extern void vga_putchar(unsigned char c);

extern void setenv(char *var, char *value);
extern void unsetenv(char *var);
extern char *getenv(char *var);
extern void printenv(void);
extern void env_init(void);

#ifndef LINUX_AT
extern int device_open(const char *name);
extern void device_close(int fd);
extern int device_read(int fd, char *buffer, int size, int offset);
#endif

extern void kbd_init(void);
extern int kbd_getc_with_timeout(int nsecs);
extern int kbd_getc(void);
extern int kbd_gets(char * charbuf, int maxlen);


extern int parse_device_name(char *name);
extern void parse_device(unsigned int device, char *name);
extern int atoi(char *s);

extern void uncompress_kernel(int fd, void *where);

extern void nvram_copyin(int offset, char * dest, int size);
extern void nvram_copyout(int offset, char * src, int size);

/********************************************************************************
 * Page frame macros                                                            *
 ********************************************************************************/
extern unsigned int allocate_pfn(int type);
unsigned int allocate_many_pfn(unsigned int count, int type);

#define FREE_PAGE 0
#define ALLOCATED_PAGE 2
#define TEMPORARY_PAGE 4

#define _ALLOCATE_PFN()            allocate_pfn(ALLOCATED_PAGE)
#define _PFN_TO_PHYSICAL(pfn)      (((unsigned long long)((pfn) << PAGE_SHIFT)) | PAGE_OFFSET)
#define _PHYSICAL_TO_PFN(physical) (((unsigned long long)(physical) & (~PAGE_OFFSET)) >> PAGE_SHIFT)

/********************************************************************************
 * General macros                                                               *
 ********************************************************************************/

#endif

