/*
 * Jacket routine for printk() in the miniloader.  It now calls vsprintf() in lib.a.
 *
 * david.rusling@reo.mts.dec.com
 */

#include <asm/system.h>
#include <linux/kernel.h>
#include <stdarg.h>
#include "miniboot.h"
#include "uart.h"

int cons_puts(int unit, char *buf);
void cons_gets(char *buf);

int printk(const char *fmt, ...)
{
    unsigned long flags;
    char buf[132];
    va_list ap;
    int count;

    save_flags(flags);
    cli();

    va_start(ap, fmt);

    count = vsprintf(buf, fmt, ap);
    cons_puts(0, buf);

    va_end(ap);				/* clean up */

    restore_flags(flags);

    return count;
}

int cons_puts(int unit, char *buf)
{
    int i;

    i = 0;
    while (buf[i] != '\0') {
	vga_putchar(buf[i]);
#ifdef MINI_SERIAL_ECHO
	uart_putchar(COM1, buf[i]);
#endif
	i++;
    }
    return i;
}

void PutChar(int c)
{
    char buf[2];
    
    buf[0] = c;
    buf[1] = '\0';
    cons_puts(0, buf);
}

void cons_gets(char *buf)
{
    int i;
    int retval;

    for (i = 0;; i++) {
	retval = uart_getchar(COM1);
	retval &= 0xff;
	switch (retval) {
	    case '\r': 
	    case '\n': 
		buf[i] = '\0';
		retval = '\r';
		uart_putchar(COM1, '\n');
		return;

	    case 0x08: 			/* BS */
	    case 0x7f: 			/* Delete */
		if (i > 0) {
		    cons_puts(0, "\b \b");
		    i--;
		}
		break;

	    default: 
		uart_putchar(COM1, retval);
		buf[i] = retval;
		break;
	}
    }
}

int GetChar(void)
{
    int c;

    c = uart_getchar(COM1);
    if (c == '\r') c = '\n';

    return c;
}
