#include <linux/config.h>
#include <linux/fs.h>
#include <linux/timer.h>
#include <linux/signal.h>
#include <linux/errno.h>
#include <linux/pci.h>
#include <linux/tqueue.h>
#include <linux/sched.h>
#include "miniboot.h"
#include "uart.h"


#undef TIMER_DEBUG 

struct timer_struct timer_table[32];
unsigned long timer_active;

struct timer_list *timers;

struct tq_struct tq_last = {
  &tq_last, 0, 0, 0
  };

DECLARE_TASK_QUEUE(tq_timer);
DECLARE_TASK_QUEUE(tq_immediate);
DECLARE_TASK_QUEUE(tq_scheduler);

int loops_per_sec = 1;

unsigned long volatile jiffies = 0;

void do_timer(int irq, struct pt_regs *regs);

/*****************************************************************************
 * initialise timers                                                         *
 *****************************************************************************/
void init_timers(void) 
{
#define TIMER_IRQ 0

    timers = NULL;
    if (request_irq(TIMER_IRQ, do_timer, 0, "timer") != 0) {
	printk("Could not allocate timer IRQ!");
	halt();
    }
}

/*****************************************************************************
 * Timer routines                                                            *
 *****************************************************************************/
void add_timer(struct timer_list *timer)
{
    unsigned long flags;

    save_flags(flags);
    cli();

#ifdef TIMER_DEBUG
    printk("add_timer called @ 0x%X with delay of %d\n", 
      timer, timer->expires);
#endif
    timer->next = timers;
    timer->prev = &timers;

    if (timers != NULL)
	timers->prev = timer;

    timers = timer;
    restore_flags(flags);
}

int del_timer(struct timer_list *timer)
{
    struct timer_list *last, *this;
    unsigned long flags;

    save_flags(flags);
    cli();
#ifdef TIMER_DEBUG
    printk("del_timer called @ 0x%X\n", timer);
#endif

    last = NULL;
    this = timers;
    while (this != NULL) {
	if (this == timer) {		/* found it        */
	    if (last == NULL)		/* 1st on list     */
		timers = this->next;
	    else			/* NOT 1st on list */
		last->next = this->next;
	    timer->next = NULL;
	    timer->prev = NULL;
	    restore_flags(flags);
	    return 1;
	}
	last = this;
	this = this->next;
    }
    timer->next = NULL;
    timer->prev = NULL;
    restore_flags(flags);
    return 1;
}

void do_timer(int irq, struct pt_regs *regs)
{
    struct timer_list *timer;
    unsigned long flags;

    save_flags(flags);
    cli();

    jiffies++;

    timer = timers;
    while (timer != NULL) {
	if (timer->expires != 0) {
	    timer->expires--;
	    if (timer->expires == 0) {
#ifdef TIMER_DEBUG
		printk("timer @ 0x%p expired\n", timer);
		printk("function @ 0x%p\n", timer->function);
		printk("PC = %016lx PS=%04lx\n", regs->pc, regs->ps);
		printk("ret = %016lx\n", regs->r26);
#endif
		(timer->function)(timer->data);
	    }
	}
	timer = timer->next;
    }

    /* now the tqueue stuff */
    run_task_queue(&tq_timer);

    restore_flags(flags);
}


