/* The following code was stolen from OSF, but it's really Thomas Roell's
 * VGA code...
 *
 * Copyright 1990,91 by Thomas Roell, Dinkelscherben, Germany.
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Thomas Roell not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  Thomas Roell makes no representations
 * about the suitability of this software for any purpose.  It is provided
 * "as is" without express or implied warranty.
 */
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/tty.h>
#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/kd.h>

#include <asm/io.h>
#include <asm/system.h>
#include <asm/segment.h>

#include "miniboot.h"

#define VGA_FB_ADDR     0xa0000
#define VGA_TX_ADDR     0xb8000

static unsigned short int *vga_tx_addr;
extern void bcopy(char *dst, char *src, int count);

/*
 *  The following two definitions control debug printing and which font to select (see
 *  the note below).
 */
#define DEBUG_PRINT 1
/* #define JDU_CONS_FONT 1 */

/* Fonts */

/*
 *  vgafonts.c is included as it contains static definitions of no interest to other code.
 *  Furthermore, I've conditionalised out the unused fonts to save space.  You switch between
 *  the two sets of fonts by defining (or not) JDU_CONS_FONT.
 */
#include "vgafonts.c"

/* Ditto for colors... this defines vga_default_colortab[] */
#include "vgacolors.c"

#define MiscOutReg General[0]

typedef struct {
    unsigned char General[4];		/* General Registers */
    unsigned char CRTC[25];		/* Crtc Controller */
    unsigned char Sequencer[5];		/* Video Sequencer */
    unsigned char Graphics[9];		/* Video Graphics */
    unsigned char Attribute[21];	/* Video Atribute */
    unsigned char DAC[768];		/* Internal Colorlookuptable */
    unsigned char NoClock;		/* number of selected clock */
    unsigned int FontInfo[2048];	/* save area for fonts */
    unsigned char TextAttribute;	/* not really a register */
} vgaHWRec, *vgaHWPtr;

#define INB(port)		inb((unsigned long)(port))
#define OUTB(port, val)		{outb_p((val), (unsigned long)(port)); mb();}
#define OUTW(port, val)         {outw((val), (unsigned long)(port)); mb();}

#define MINW(port)              __readw((unsigned long)(port))
#define MOUTW(port, val)         {__writew((val), (unsigned long)(port)); mb();}

int vga_delay = 2;			/* 200000 */
#define VGA_DELAY()                             \
  {                                             \
    volatile int i;                             \
    for (i = vga_delay; i; i--);                \
  }

/* forward declarations */
void vga_clearscreen(void);
void vga_scrollscreen(void);

#if 1
/* ------------------------------------------------------------------ */
/*
 *  Number 9 specific code.
 */
/* ------------------------------------------------------------------ */
int locate_number9(void) 
{

    unsigned char iotemp;
    int status;

    /* Unlock the S3 registers. Try both. */

    OUTB(0x03D4, 0x38);
    OUTB(0x03D5, 0x48);

    OUTB(0x03D4, 0x39);
    OUTB(0x03D5, 0xa5);

    OUTB(0x03D4, 0x30);
    iotemp = (unsigned char) INB(0x03D5);
    iotemp = iotemp & 0xFF;

    if ((iotemp > 0x93) && (iotemp < 0x99))
	status = 1;
    else
	status = 0;

#ifdef DEBUG_PRINT
    printk("In locate_number9, ID = %x, status = %x\n", iotemp, status);
#endif

    return (status);
}
#endif

/************************************************************************/
/*									*/
/*    Code from OSF snapshot - vga.c					*/
/*									*/
/************************************************************************/

/*
 *  VGA Device Driver
 *  
 *  Created 4/91 by Dean Anderson
 *
 */
/*
 * Copyright 1990,91 by Thomas Roell, Dinkelscherben, Germany.
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Thomas Roell not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  Thomas Roell makes no representations
 * about the suitability of this software for any purpose.  It is provided
 * "as is" without express or implied warranty.
 */


int reverseByte(int b)
{
    register int j, k, t = 0;

    for (j = 0x01, k = 0x80; j < k; j += j, k >>= 1) {
	if (b & j)
	    t |= k;
	if (b & k)
	    t |= j;
    }

    return (t);

}
/*
 * routine to place the default font information into the state record
 */
void vga_copy_default_font(vgaHWRec *new)
{
    register int i;

#ifndef JDU_CONS_FONT
    register int j;

#endif
    register unsigned char *src, *dst;

#ifdef DEBUG_PRINT
    printk("vga_copy_default_font: entry\n");
#endif					/* DEBUG_PRINT */

    dst = (unsigned char *) &new->FontInfo[0];

    /* Load in the character fonts for the full range of characters which is
     * 0x0-0xFF (0-255). */
    for (i = 0; i < 256; ++i) {

	/* Initially specify the font for a "bogus" character.  This is used
	 * for non-printable characters.  It is an all F's glyph which should
	 * make it obvious on the display if you are attempting to display an
	 * invalid character. */
	src = &vga8x16xx[0];

	/* This locates the character glyph for the conventional lower range
	 * 7-bit printable ascii characters.  This is for the space character
	 * up to the DEL (0177) character. */
	if (i >= 0x20 && i < 0x7F) {	/* Figure out the offset into the set
					 * of character glyphs. This is done by
					 * subtracting the char code (i) from
					 * the first printable char code (0x20)
					 * to get the offset into the array.
					 * Then multiply by 16 because each
					 * character consists of 16 bytes; this
					 * computes the address of the glyph
					 * for (i). */
#ifdef JDU_CONS_FONT
	    src = &vga8x16gl[16 * (i - 0x20)];
#else
	    src = &q_font[15 * (i - 0x20)];
					/* q_font is 8x15 */
#endif					/* JDU_CONS_FONT */
	}

	/* This sets a pointer to the character glyph for characters in upper
	 * range of printable characters.  These consist mostly of accent typed
	 * characters used in non-English. */
#ifdef JDU_CONS_FONT
	else if (i >= 0xA0 && i <= 0xFF) {
	    src = &vga8x16gr[16 * (i - 0xA0)];
	}
#else
	else if (i >= 0xA1 && i <= 0xFF) {
	    src = &q_font[15 * (i - 0x42)];
					/* q_font is 8x15 */
	}
#endif					/* JDU_CONS_FONT */

	/* At this point "src" points to the beginning byte offset for the
	 * glyph representing the character "i". Copy the 16 bytes used to
	 * represent that character out to the destination area. */
#ifdef JDU_CONS_FONT
	bcopy(src, dst, 16);
	dst += 16;
#else
/* FIXME FIXME - font from qfont.c is bit-reversed here... :-( :-( */
	for (j = 0; j < 15; j++) {
	    *dst++ = reverseByte(*src++);
	}
	*dst++ = 0;
#endif					/* JDU_CONS_FONT */

	/* 16 bytes currently unused per 32-byte char entry */
	bcopy(vga8x16unused, dst, 16);
	dst += 16;
    }
}
#ifdef DEBUG_PRINT
/*
 *-----------------------------------------------------------------------
 * vgaHWPrint --
 *      Print contents of VGA HW regsiters
 *
 * Input:
 *      pointer to the mode record.
 *
 * Results:
 *	new mode record contents printed.
 *
 * Side Effects: 
 *
 *-----------------------------------------------------------------------
 */
void vgaHWPrint(vgaHWPtr tmp)
{
    int i;

    printk("\nClock number (0x%lx) 0x%x\n", &tmp->NoClock, tmp->NoClock);
    printk("\nGeneral Registers: (0x%lx)\n  ", tmp->General);
    for (i = 0; i < 4; i++)
	printk("0x%x ", tmp->General[i]);

    printk("\nSequencer Registers: (0x%lx)\n  ", tmp->Sequencer);
    for (i = 0; i < 5; i++)
	printk("0x%x  ", tmp->Sequencer[i]);

    printk("\nCRTC Registers: (0x%lx)\n  ", tmp->CRTC);
    for (i = 0; i < 8; i++)
	printk("0x%x  ", tmp->CRTC[i]);
    printk("\n  ");
    for (i = 8; i < 16; i++)
	printk("0x%x  ", tmp->CRTC[i]);
    printk("\n  ");
    for (i = 16; i < 25; i++)
	printk("0x%x  ", tmp->CRTC[i]);


    printk("\nGraphics Registers: (0x%lx)\n  ", tmp->Graphics);
    for (i = 0; i < 9; i++)
	printk("0x%x  ", tmp->Graphics[i]);

    printk("\nAttribute Registers: (0x%lx)\n  ", tmp->Attribute);
    for (i = 0; i < 8; i++)
	printk("0x%x  ", tmp->Attribute[i]);
    printk("\n  ");
    for (i = 8; i < 16; i++)
	printk("0x%x  ", tmp->Attribute[i]);
    printk("\n  ");
    for (i = 16; i < 21; i++)
	printk("0x%x  ", tmp->Attribute[i]);
    printk("\n");

    printk("\n Text Attribute 0x%x\n", tmp->TextAttribute);

    locate_number9();
    OUTB(0x3d4, 0x30)
}
#endif

/*
 *-----------------------------------------------------------------------
 * vgaHWInit --
 *      Handle the initialization, etc. of a screen.
 *
 * Results:
 *      None.
 *
 * Side Effects: 
 *
 *-----------------------------------------------------------------------
 */

void vgaHWInit(vgaHWRec *new)
{
    int i;

#ifdef DEBUG_PRINT
    printk("vgaHWInit: entry\n");
#endif					/* DEBUG_PRINT */

    /* initialize default colormap */
    for(i = 0; i < 768; i++) {
	new->DAC[i] = vga_default_colortab[i];
    }

/* FIXME FIXME - init like JDU console */


/* FIXME FIXME - replaced algorithmically-determined register contents */
/* FIXME FIXME - code from OSF with hardwired values based on BOOK and */
/* FIXME FIXME - JDU console driver, for 80x25 text mode */
/* FIXME FIXME - This means that the state should be ready for chars */
/* FIXME FIXME - to be written into the "screen" memory at b8000 */

    new->MiscOutReg = 0x67;		/* General[0] , 0x3c2, color, cpu
					 * access to video memory, 28 Mhz clock
					 */
    new->General[1] = 0;		/* 0x3ca, EGA only */
    new->General[2] = 0x10;		/* 0x3c2, MiscOut, write */
    new->General[3] = 0xc5;		/* 0x3da, Feature control */

    /* Time Sequencer (a = 0x3c4, d = 0x3c5) */
    new->Sequencer[0] = 0x00;		/* synchronous reset, ignored by the
					 * code */
#if 0
    new->Sequencer[1] = 0x01;		/* character clocks are 8 dots wide */
    new->Sequencer[2] = 0x03;		/* enables writes to planes 0 and 1 */
    new->Sequencer[3] = 0x00;		/* Font select */
    new->Sequencer[4] = 0x03;		/* Memory mode, bit 0 = 1 = character
					 * mode, bit 1 = 1 = more than 64k
					 * bytes on adaptor */
#else
    new->Sequencer[1] = 0x00;		/* character clocks are 9 dots wide */
    new->Sequencer[2] = 0x03;		/* Map mask */
    new->Sequencer[3] = 0x00;		/* font select ( = 0 ) */
    new->Sequencer[4] = 0x03;		/* Memory mode */
#endif


    /* CRTC Controller (a = 0x3d4, d = 0x3d5) */
    new->CRTC[0] = 0x5f;		/* total character clocks - 5 */
    new->CRTC[1] = 0x4f;		/* number character clocks displayed - 
					 * 1 */
    new->CRTC[2] = 0x50;
    new->CRTC[3] = 0x82;
    new->CRTC[4] = 0x55;
    new->CRTC[5] = 0x81;
    new->CRTC[6] = 0xbf;
    new->CRTC[7] = 0x1f;
    new->CRTC[8] = 0x00;
    new->CRTC[9] = 0x4f;		/* DOC 0xc7 */
    new->CRTC[10] = 0x01;		/* DOC 0x06 */
					/* cursor start */
    new->CRTC[11] = 0x0f;		/* DOC 0x07 */
					/* cursor end */
    new->CRTC[12] = 0x00;		/* start addr high */
    new->CRTC[13] = 0x00;		/* start addr low */
    new->CRTC[14] = 0x00;
    new->CRTC[15] = 0x00;		/* DOC 0x59 */
    new->CRTC[16] = 0x9c;
    new->CRTC[17] = 0x3e;		/* unlock CRTC regs 0-7, Vert int
					 * disable/normal */
    new->CRTC[18] = 0x8f;
    new->CRTC[19] = 0x28;
    new->CRTC[20] = 0x1f;
    new->CRTC[21] = 0x96;
    new->CRTC[22] = 0xb9;
    new->CRTC[23] = 0xa3;
    new->CRTC[24] = 0xff;

    /* Graphics Display Controller (a = 0x3ce, d = 0x3cf) */
    new->Graphics[0] = 0x00;
    new->Graphics[1] = 0x00;
    new->Graphics[2] = 0x00;
    new->Graphics[3] = 0x00;
    new->Graphics[4] = 0x00;
    new->Graphics[5] = 0x10;
    new->Graphics[6] = 0x0e;
    new->Graphics[7] = 0x00;
    new->Graphics[8] = 0xff;

    /* Attributes (a = 0x3c0, d = 0x3c0) */
    new->Attribute[0] = 0x00;		/* standard colormap translation */
    new->Attribute[1] = 0x01;
    new->Attribute[2] = 0x02;
    new->Attribute[3] = 0x03;
    new->Attribute[4] = 0x04;
    new->Attribute[5] = 0x05;
    new->Attribute[6] = 0x06;
    new->Attribute[7] = 0x07;
    new->Attribute[8] = 0x08;
    new->Attribute[9] = 0x09;
    new->Attribute[10] = 0x0a;
    new->Attribute[11] = 0x0b;
    new->Attribute[12] = 0x0c;
    new->Attribute[13] = 0x0d;
    new->Attribute[14] = 0x0e;
    new->Attribute[15] = 0x0f;
#if 1
    new->Attribute[16] = 0x00;		/* Not graphics mode, ie alphanumeric
					 */
    new->Attribute[17] = 0x0c;		/* screen border */
    new->Attribute[18] = 0x0f;		/* Color plane enable, planes 3:0 are
					 * enabled */
    new->Attribute[19] = 0x00;		/* Horizontal PEL planing */
    new->Attribute[20] = 0x00;		/* Color select, first palette */
#else
    new->Attribute[16] = 0x8c;		/* Not graphics mode, ie alphanumeric
					 */
    new->Attribute[17] = 0x00;		/* screen border */
    new->Attribute[18] = 0x0f;		/* Color plane enable, planes 3:0 are
					 * enabled */
    new->Attribute[19] = 0x08;		/* Horizontal PEL planing */
    new->Attribute[20] = 0x00;		/* Color select, first palette */
#endif

    /* now, copy the font information into its area */
    vga_copy_default_font(new);

}


/*
 *-----------------------------------------------------------------------
 * vgaHWRestore --
 *      restore a video mode
 *
 * Results:
 *      nope.
 *
 * Side Effects: 
 *      the display enters a new graphics mode. 
 *
 *-----------------------------------------------------------------------
 */
void vgaHWRestore(vgaHWPtr restore)
{
    register int i;
    register unsigned short *fp;
    void *dstp;
    int data;

#ifdef DEBUG_PRINT
    printk("vgaHWRestore: entry\n");
#endif					/* DEBUG_PRINT */

    OUTB(0x3C2, restore->MiscOutReg);	/* General[0] */

    OUTB(0x3D4, 0x11);
    OUTB(0x3D5, 0x00);			/* unlock CRTC registers 0-7 */

/* FIXME FIXME ?? - kinda mucked this about a bit for the Sequencer regs... */
    OUTB(0x3C4, 0x00);			/* select Reset */
    OUTB(0x3C5, 0x01);			/* Bit 0 = 1 = VGA Syncronous reset */
    for (i = 1; i < 5; i++) {
	OUTB(0x3C4, i);
	OUTB(0x3C5, restore->Sequencer[i]);
    }
    OUTB(0x3C4, 0x00);			/* Select Reset */
    OUTB(0x3C5, 0x03);			/* Bit 0 = VGA Synchronous reset, Bit 1
					 * =  now reenable the timing sequencer
					 */

    data = INB(0x3DA);			/* reset flip-flop */
    for (i = 0; i < 16; i++) {
	OUTB(0x3C0, i);
	OUTB(0x3C0, restore->Attribute[i]);
    }
    for (i = 16; i < 21; i++) {
	OUTB(0x3C0, i);
	OUTB(0x3C0, restore->Attribute[i]);
    }
    OUTB(0x3c0, 0x20);			/* FIXME - 'cuz JDU does it this way */

    for (i = 0; i < 25; i++) {
	OUTB(0x3D4, i);
	OUTB(0x3D5, restore->CRTC[i]);
    }

    for (i = 0; i < 9; i++) {
	OUTB(0x3CE, i);
	OUTB(0x3CF, restore->Graphics[i]);
    }

    OUTB(0x3C6, 0xff);			/* FIXME - added 'cuz in JDU code ???
					 */

    OUTB(0x3C8, 0x00);
    for (i = 0; i < 768; i++)
	OUTB(0x3C9, restore->DAC[i]);

/* FIXME FIXME - moved this to *AFTER* the regs are restored ??? */

    /* We MUST explicitly restore the font, since we (may have) entered
     *  graphics mode, which could have destroyed it. */
    if (((restore->Attribute[16] & 0x01) == 0)) {

	/* here we switch temporarily to 16 color-plane-mode, to simply copy
	 * the font-info BUGALLERT: The vga's segment-select register MUST be  
	 * set appropriate ! */
#ifdef FIXME
	data = INB(0x3da);		/* reset flip-flop */
	OUTB(0x3C0, 0x10);
	OUTB(0x3C0, 0x01);		/* graphics mode */
#endif					/* FIXME */

	/* Sequencer registers */
	OUTB(0x3C4, 0x02);
	OUTB(0x3C5, 0x04);		/* write to plane 2 */
	OUTB(0x3C4, 0x04);
	OUTB(0x3C5, 0x06);		/* enable plane graphics */

	/* Graphics registers */
#ifdef FIXME
	OUTB(0x3CE, 0x04);
	OUTB(0x3CF, 0x02);		/* read plane 2 */
#endif					/* FIXME */
	OUTB(0x3CE, 0x05);
	OUTB(0x3CF, 0x00);		/* write mode 0, read mode 0 */
	OUTB(0x3CE, 0x06);
	OUTB(0x3CF, 0x05);		/* set graphics */

	VGA_DELAY();

	fp = (unsigned short *) restore->FontInfo;
	dstp = (void *) VGA_FB_ADDR;
	for (i = 0; i < 4096; i++, dstp += sizeof(short)) {
	    MOUTW(dstp, *fp++);
	}
	mb();

#ifdef FIXME
	data = INB(0x3da);		/* reset flip-flop */
	OUTB(0x3C0, 0x10);
	OUTB(0x3C0, restore->Attribute[0x10]);
#endif					/* FIXME */

	/* Sequencer registers */
	OUTB(0x3C4, 0x02);
	OUTB(0x3C5, restore->Sequencer[2]);
	OUTB(0x3C4, 0x04);
	OUTB(0x3C5, restore->Sequencer[4]);

	/* Graphics registers */
#ifdef FIXME
	OUTB(0x3CE, 0x04);
	OUTB(0x3CF, restore->Graphics[4]);
#endif					/* FIXME */
	OUTB(0x3CE, 0x05);
	OUTB(0x3CF, restore->Graphics[5]);
	OUTB(0x3CE, 0x06);
	OUTB(0x3CF, restore->Graphics[6]);
    }

    VGA_DELAY();

}

#ifdef DEBUG_PRINT
/*
 *-----------------------------------------------------------------------
 * vgaHWSave --
 *      save the current video mode
 *
 * Results:
 *      pointer to the current mode record.
 *
 * Side Effects: 
 *      None.
 *-----------------------------------------------------------------------
 */

void vgaHWSave(vgaHWPtr save)
{
    register int i;
    int data;

#ifdef FONT_SAVE_ENABLED
    register unsigned int *fp;
    void *srcp;

#endif

#ifdef DEBUG_PRINT
    printk("vgaHWSave: entry\n");
#endif					/* DEBUG_PRINT */

    /* Here we are, when we first save the videostate. This means we came here 
     * to save the original Text (??) mode. Because some drivers may depend on 
     * NoClock we set it here to a reasonable value. */
    data = INB(0x3CC);
    save->NoClock = (data >> 2) & 3;

    /* now get the registers */
    save->MiscOutReg = INB(0x3CC);
    save->General[1] = INB(0x3CA);
    save->General[2] = INB(0x3C2);
    save->General[3] = INB(0x3DA);

    for (i = 0; i < 25; i++) {
	OUTB(0x3D4, i);
	save->CRTC[i] = INB(0x3D5);
    }

    for (i = 0; i < 5; i++) {
	OUTB(0x3C4, i);
	save->Sequencer[i] = INB(0x3C5);
    }
    for (i = 0; i < 9; i++) {
	OUTB(0x3CE, i);
	save->Graphics[i] = INB(0x3CF);
    }

    data = INB(0x3DA);			/* reset flip-flop */
    for (i = 0; i < 16; i++) {
	OUTB(0x3C0, i);
	save->Attribute[i] = INB(0x3C1);
	OUTB(0x3C0, save->Attribute[i]);
    }
    for (i = 16; i < 21; i++) {
	OUTB(0x3C0, i | 0x20);
	save->Attribute[i] = INB(0x3C1);
	OUTB(0x3C0, save->Attribute[i]);
    }
    save->CRTC[17] &= 0x7F;		/* clear protect bit !!! */

    /* assume we're still in "text" mode, so go out and read the  first
     * attribute from the text area, and squirrel it away... as a side effect, 
     * set vga_tx_addr to point to where the base address  of the current text 
     * window will be... */
    vga_tx_addr = (unsigned short int *) VGA_TX_ADDR;
    data = MINW(vga_tx_addr);
    save->TextAttribute = (data >> 8) & 0xff;

    /*                      save the colorlookuptable */
    OUTB(0x3C6, 0xff);
    OUTB(0x3C7, 0x00);
    for (i = 0; i < 768; i++)
	save->DAC[i] = INB(0x3C9);

#ifdef FONT_SAVE_ENABLED

    /* get the character set of the first non-graphics application */
    if (((save->Attribute[0x10] & 0x01) == 0)) {

	/* Here we switch temporarily to 16 color-plane-mode, to simply copy
	 * the font-info BUGALLERT: The vga's segment-select register MUST be  
	 * set appropriate ! */
	data = INB(0x3DA);		/* reset Attribute flip-flop */
	OUTB(0x3C0, 0x10);
	OUTB(0x3C0, 0x01);		/* graphics mode */

	/* Sequencer registers */
	OUTB(0x3C4, 0x02);
	OUTB(0x3C5, 0x04);		/* write to plane 2 */
	OUTB(0x3C4, 0x04);
	OUTB(0x3C5, 0x06);		/* enable plane graphics */

	/* Graphics registers */
	OUTB(0x3CE, 0x04);
	OUTB(0x3CF, 0x02);		/* read plane 2 */
	OUTB(0x3CE, 0x05);
	OUTB(0x3CF, 0x00);		/* write mode 0, read mode 0 */
	OUTB(0x3CE, 0x06);
	OUTB(0x3CF, 0x05);		/* set graphics */

	VGA_DELAY();

	fp = save->FontInfo;
	srcp = VGA_FB_ADDR;
	for (i = 0; i < 2048; i++, srcp += sizeof(int)) {
	    *fp++ = MINDW(srcp);
	}

	/* Now set things back, so that save doesn't destroy as well */
	data = INB(0x3DA);
	OUTB(0x3C0, 0x10);
	OUTB(0x3C0, save->Attribute[0x10]);

	/* Sequencer registers */
	OUTB(0x3C4, 0x02);
	OUTB(0x3C5, save->Sequencer[0x2]);
	OUTB(0x3C4, 0x04);
	OUTB(0x3C5, save->Sequencer[0x4]);

	/* Graphics registers */
	OUTB(0x3CE, 0x04);
	OUTB(0x3CF, save->Graphics[0x4]);
	OUTB(0x3CE, 0x05);
	OUTB(0x3CF, save->Graphics[0x5]);
	OUTB(0x3CE, 0x06);
	OUTB(0x3CF, save->Graphics[0x6]);

	VGA_DELAY();
    }
#endif					/* FONT_SAVE_ENABLED */

    return;
}
#endif

void vga_video_on(void) 
{
    unsigned char state;

#ifdef DEBUG_PRINT
    printk("vga_video_on: entry\n");
#endif					/* DEBUG_PRINT */

    OUTB(0x3C4, 0x01);
    state = INB(0x3C5);

    state &= 0xDF;

    /* turn on video in Clocking Mode register */
    OUTB(0x3C4, 0x01);
    OUTB(0x3C5, state);			/* change mode */
}
/*
 *-----------------------------------------------------------------------
 * vga_init --
 *      Handle the initialization of VGA register structure.
 *
 * Input:
 *      None
 *
 * Results:
 *      VGA inititalised.
 *
 * Side Effects: 
 *
 *      None.
 *
 *-----------------------------------------------------------------------
 */
static vgaHWRec new;
#ifdef DEBUG_PRINT
static vgaHWRec old;
#endif

int vga_initialised = 0;
static int column, row;

void vga_init(void) 
{

#ifdef DEBUG_PRINT
    printk("vga_init: entry\n");
#endif					/* DEBUG_PRINT */

    /* setup the default values for the device */
    vgaHWInit(&new);

    /* check to see if the card is enabled, if not, then enable it */
printk("checking if VGA card is enabled\n");
    OUTB(0x3c2, 0x01);
    if (INB(0x3cc) != 0x01) {
printk(" ... NOPE, enabling it\n");
	outb(0x10, 0x46E8);		/* video enable.  Bit 4: enables r/w to
					 * port 102 */
	outb(0x01, 0x0102);		/* Options select register.   Respond
					 * to VGA addresses. */
	outb(0x08, 0x46E8);		/* video enable.  Bit 3: video
					 * subsystem enable */
    }
OUTB(0x3c2, 0x01);
printk(" ... 3c2==0x%x\n", INB(0x3cc));

#ifdef DEBUG_PRINT

    /* get the current setup and display it */
    vgaHWSave(&old);
    vgaHWPrint(&old);
#endif

    /* reset to the new set up */
    vgaHWRestore(&new);

    /* tell the world */
#ifdef DEBUG_PRINT
    vgaHWPrint(&new);
#endif

    /* turn the video on */
    vga_video_on();

    /* variables used by vga_putchar() */
    vga_tx_addr = VGA_TX_ADDR;
    row = column = 0;

    /* clear the screen */
    vga_clearscreen();

    vga_initialised = 1;

}

/*
 *-----------------------------------------------------------------------
 * vga_putchar --
 *      Put a character out to screen memory.
 *
 *      The screen is 80*25
 *-----------------------------------------------------------------------
 */
void vga_putchar(unsigned char c)
{
    if (vga_initialised) {
	int		cursorloc;
	unsigned short int outword;

	if (c == '\n') {
	    /* just go onto the next row */
	    row++;
	    column = 0;
	} 
	else if(c == '\r') {
	    /* Go to beginning of current row */
	    column = 0;
	}
	else if(c == '\b') {
	    /* Backspace: go back one column unless we're already at
	     * beginning of row...
	     */
	    if(column > 0) column--;
	}
	else {
	    /* build the character */
	    outword = (unsigned short int) c;
	    outword = outword | (0x07 << 8);

	    /* output it to the correct location */
	    vga_tx_addr = (unsigned short *)
		(VGA_TX_ADDR + (2 * row * 80) + (2 * column));
	    MOUTW(vga_tx_addr, outword);

	    /* move on where to write the next character */
	    column++;
	}
	/* check for rows and columns overflowing */
	if (column > 79) {
	    column = 0; row++;
	}
	if (row > 24) {
	    vga_scrollscreen();
	}

	/* Put the cursor at the right spot... */
	cursorloc = ((row * 80) + column);
	OUTB(0x3d4, 14);
	OUTB(0x3d5, (cursorloc & 0xff00) >> 8);
	OUTB(0x3d4, 15);
	OUTB(0x3d5, (cursorloc & 0xff));

    }
}

/*
 *-----------------------------------------------------------------------
 * vga_clearscreen --
 *      clear the screen by writing spaces to every location
 *      The screen is 80*25
 *-----------------------------------------------------------------------
 */
void vga_clearscreen(void)
{
    int i;
    unsigned short int outword;

    outword = (unsigned short int) ' ';
    outword = outword | (0x07 << 8);
    vga_tx_addr = VGA_TX_ADDR;

    for (i = 0; i < (25 * 80); i++) {
	MOUTW(vga_tx_addr, outword);
	vga_tx_addr++;
    }

    /* reset to the first location */
    row = column = 0;
}

void vga_scrollscreen(void)
{
    int i;
    unsigned short int	ioword;
    unsigned short int *source_tx_addr;

    /* Scroll the screen by copying the entire text buffer up
     * one line (80 chars).
     */
    vga_tx_addr = VGA_TX_ADDR;
    source_tx_addr = vga_tx_addr + 80;

    for (i = 0; i < (24*80); i++) {
	ioword = MINW(source_tx_addr);
	MOUTW(vga_tx_addr, ioword);
	vga_tx_addr++; source_tx_addr++;
    }

    /* Clear the last row */
    ioword = (unsigned short int) ' ';
    ioword = ioword | (0x07 << 8);

    for(i = 0; i < 80; i++) {
	MOUTW(vga_tx_addr, ioword);
	vga_tx_addr++;
    }

    /* Position cursor at bottom of cleared row */
    row = 24;
    column = 0;
    vga_tx_addr -= 80;
}
